import logging

import numpy as np
import torch
from torch.nn import CrossEntropyLoss
from torch.optim import SGD, Adam
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import ListOfBlobs, Session
from idtrackerai.utils import conf, load_id_images

from ..network import (
    CNN,
    DEVICE,
    LearnerClassification,
    NetworkParams,
    StopTraining,
    get_dataloader,
    get_predictions,
    train_loop,
)
from .crossings_dataset import get_train_validation_and_eval_blobs
from .model_area import ModelArea


def apply_area_and_unicity_heuristics(list_of_blobs: ListOfBlobs, n_animals: int):
    logging.info(
        "Classifying Blobs as individuals or crossings "
        "depending on their area and the number of blobs in the frame"
    )

    model_area = ModelArea(list_of_blobs, n_animals)

    for blobs_in_frame in list_of_blobs.blobs_in_video:
        unicity_cond = len(blobs_in_frame) == n_animals
        for blob in blobs_in_frame:
            blob.seems_like_individual = unicity_cond or model_area(blob.area)


def detect_crossings(list_of_blobs: ListOfBlobs, session: Session):
    """Classify all blobs in the video as being crossings or individuals"""

    apply_area_and_unicity_heuristics(list_of_blobs, session.n_animals)

    train_images, train_labels, train_weights, val_images, val_labels = (
        get_train_validation_and_eval_blobs(
            list_of_blobs.blobs_in_video, session.n_animals
        )
    )

    unknown_blobs = [
        blob
        for blob in list_of_blobs.all_blobs
        if not hasattr(blob, "is_an_individual")
    ]

    logging.info(f"{len(unknown_blobs)} unknown blobs")

    if (
        np.count_nonzero(train_labels)
        < conf.MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR
    ):
        logging.debug("There are not enough crossings to train the crossing detector")
        for blob in unknown_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return
    logging.info("There are enough crossings to train the crossing detector")

    train_loader = get_dataloader(
        "training",
        load_id_images(session.id_images_file_paths, train_images),
        train_labels,
        conf.BATCH_SIZE_DCD,
    )

    val_loader = get_dataloader(
        "validation",
        load_id_images(session.id_images_file_paths, val_images),
        val_labels,
    )

    logging.info("Setting crossing detector network parameters")
    network_params = NetworkParams(
        n_classes=2,
        architecture="CNN",
        save_folder=session.crossings_detector_folder,
        model_name="crossing_detector",
        image_size=session.id_image_size,
        optimizer="Adam",
        schedule=[30, 60],
        optim_args={"lr": conf.LEARNING_RATE_DCD},
        epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_DCD,
    )
    network_params.save()

    crossing_model = CNN.from_network_params(network_params).to(DEVICE)

    if network_params.optimizer == "Adam":
        optimizer = Adam(crossing_model.parameters(), **network_params.optim_args)
    elif network_params.optimizer == "SGD":
        optimizer = SGD(crossing_model.parameters(), **network_params.optim_args)
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)
    criterion = CrossEntropyLoss(
        weight=torch.tensor(train_weights, dtype=torch.float32)
    ).to(DEVICE)
    learner = LearnerClassification(crossing_model, criterion, optimizer, scheduler)
    stopping = StopTraining(
        epochs_limit=conf.MAXIMUM_NUMBER_OF_EPOCHS_DCD,
        overfitting_limit=conf.OVERFITTING_COUNTER_THRESHOLD_DCD,
        plateau_limit=conf.LEARNING_RATIO_DIFFERENCE_DCD,
    )

    try:
        train_loop(learner, train_loader, val_loader, stopping)
    except RuntimeError as exc:
        logging.warning(
            "[red]The model diverged[/] provably due to a bad segmentation. Falling"
            " back to individual-crossing discrimination by average area model."
            " Original error: %s",
            exc,
            extra={"markup": True},
        )
        for blob in unknown_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return

    del train_loader
    del val_loader

    learner.save_model(network_params.model_path)
    logging.info("Using crossing detector to classify individuals and crossings")
    predictions, _softmax = get_predictions(
        crossing_model,
        [(blob.id_image_index, blob.episode) for blob in unknown_blobs],
        session.id_images_file_paths,
        "crossings",
    )

    logging.info(
        "Prediction results: %d individuals and %d crossings",
        np.count_nonzero(predictions == 1),
        np.count_nonzero(predictions == 2),
    )
    for blob, prediction in zip(unknown_blobs, predictions):
        blob.is_an_individual = prediction != 2

    list_of_blobs.update_id_image_dataset_with_crossings(session.id_images_file_paths)



import logging
import shutil
import sys
from argparse import ArgumentError
from importlib.metadata import version
from importlib.resources import files
from pathlib import Path
from typing import Any

from idtrackerai import Session
from idtrackerai.utils import (
    IdtrackeraiError,
    conf,
    load_toml,
    manage_exception,
    pprint_dict,
    wrap_entrypoint,
)

from .arg_parser import parse_args


def gather_input_parameters() -> tuple[bool, dict[str, Any]]:
    parameters = {}
    if Path("local_settings.py").is_file():
        logging.warning("Deprecated local_settings format found in ./local_settings.py")

    local_settings_path = Path("local_settings.toml")
    if local_settings_path.is_file():
        parameters = load_toml(local_settings_path)

    try:
        terminal_args = parse_args()
    except ArgumentError as exc:
        raise IdtrackeraiError() from exc

    ready_to_track = terminal_args.pop("track")

    if "general_settings" in terminal_args:
        parameters.update(load_toml(terminal_args.pop("general_settings")))
        logging.warning(
            "The terminal argument --settings is deprecated, please use --load with"
            " multiple files instead."
        )

    if "parameters" in terminal_args:
        for parameter_file in terminal_args.pop("parameters"):
            parameters.update(load_toml(parameter_file))
    else:
        logging.info("No parameter files detected")

    if terminal_args:
        logging.info(
            pprint_dict(terminal_args, "Terminal arguments"), extra={"markup": True}
        )
        parameters.update(terminal_args)
    else:
        logging.info("No terminal arguments detected")
    return ready_to_track, parameters


@wrap_entrypoint
def main() -> bool:
    """The command `idtrackerai` runs this function"""
    ready_to_track, user_parameters = gather_input_parameters()

    session = Session()
    non_recognized_params_1 = conf.set_parameters(**user_parameters)
    non_recognized_params_2 = session.set_parameters(**user_parameters)

    non_recognized_params = non_recognized_params_1 & non_recognized_params_2

    if non_recognized_params:
        raise IdtrackeraiError(f"Not recognized parameters: {non_recognized_params}")

    if not ready_to_track:
        ready_to_track = run_segmentation_GUI(session)
        if not ready_to_track:
            return False

    from idtrackerai.base.run import RunIdTrackerAi

    return RunIdTrackerAi(session).track_video()


def run_segmentation_GUI(session: Session | None) -> bool:
    try:
        from qtpy.QtWidgets import QApplication

        from idtrackerai.segmentation_app import SegmentationGUI
    except ImportError as exc:
        raise IdtrackeraiError(
            "\n\tRUNNING AN IDTRACKER.AI INSTALLATION WITHOUT ANY QT BINDING.\n\tGUIs"
            " are not available, only tracking directly from the terminal with the"
            " `--track` flag.\n\tRun `pip install pyqt5` or `pip install pyqt6` to"
            " build a Qt binding"
        ) from exc

    # this catches exceptions when raised inside Qt
    def excepthook(exc_type, exc_value, exc_tb):
        assert QApplication  # Pylance is happier with this
        QApplication.quit()
        manage_exception(exc_value)

    sys.excepthook = excepthook
    app = QApplication(sys.argv)
    signal = {"run_idtrackerai": False}
    window = SegmentationGUI(session, signal)
    window.show()
    app.exec()
    return signal["run_idtrackerai"] is True


@wrap_entrypoint
def general_test():
    from datetime import datetime

    COMPRESSED_VIDEO_PATH = Path(str(files("idtrackerai"))) / "data" / "test_B.avi"

    video_path = Path.cwd() / COMPRESSED_VIDEO_PATH.name
    shutil.copyfile(COMPRESSED_VIDEO_PATH, video_path)

    session = Session()
    session.set_parameters(
        name="test",
        video_paths=video_path,
        tracking_intervals=None,
        intensity_ths=[0, 130],
        area_ths=[150, float("inf")],
        number_of_animals=8,
        resolution_reduction=1.0,
        check_segmentation=False,
        ROI_list=None,
        track_wo_identities=False,
        use_bkg=False,
        protocol3_action="continue",
    )

    _ready_to_track, user_parameters = gather_input_parameters()
    non_recognized_params_1 = conf.set_parameters(**user_parameters)
    non_recognized_params_2 = session.set_parameters(**user_parameters)
    non_recognized_params = non_recognized_params_1 & non_recognized_params_2
    if non_recognized_params:
        raise IdtrackeraiError(f"Not recognized parameters: {non_recognized_params}")
    from idtrackerai.base.run import RunIdTrackerAi

    start = datetime.now()
    success = RunIdTrackerAi(session).track_video()
    if success:
        logging.info(
            "[green]Test passed successfully in %s with version %s",
            str(datetime.now() - start).split(".")[0],
            version("idtrackerai"),
            extra={"markup": True},
        )


if __name__ == "__main__":
    main()



from typing import Callable

import numpy as np

from idtrackerai import Blob, Fragment, Session
from idtrackerai.utils import track


def produce_trajectories(
    blobs_in_video: list[list[Blob]],
    number_of_animals: int,
    progress_bar=None,
    abort: Callable = lambda: False,
    fragments: list[Fragment] | None = None,
):
    """Produce trajectories array from ListOfBlobs

    Parameters
    ----------
    blobs_in_video : <ListOfBlobs object>
        See :class:`list_of_blobs.ListOfBlobs`
    number_of_frames : int
        Total number of frames in video
    number_of_animals : int
        Number of animals to be tracked

    Returns
    -------
    dict
        Dictionary with np.array as values (trajectories organized by identity)

    """
    number_of_frames = len(blobs_in_video)
    centroid_trajectories = np.full((number_of_frames, number_of_animals, 2), np.NaN)
    id_probabilities = np.full((number_of_frames, number_of_animals, 1), np.NaN)
    areas = np.full((number_of_frames, number_of_animals), np.NaN)

    for frame_number, blobs_in_frame in enumerate(
        track(blobs_in_video, "Producing trajectories")
    ):
        if abort():
            return None, None, {}
        if progress_bar:
            progress_bar.emit(frame_number)
        for blob in blobs_in_frame:
            for identity, centroid in blob.final_ids_and_centroids:
                if identity not in (None, 0):
                    centroid_trajectories[blob.frame_number, identity - 1] = centroid
            blob_final_identities = list(blob.final_identities)
            if blob.is_an_individual and len(blob_final_identities) == 1:
                identity = blob_final_identities[0]
                if identity in (None, 0):
                    continue

                areas[blob.frame_number, identity - 1] = blob.area

                if fragments is None:
                    continue
                P2_vector = fragments[blob.fragment_identifier].P2_vector

                if P2_vector is None:
                    continue
                id_probabilities[blob.frame_number, identity - 1] = P2_vector.max()

    return (
        centroid_trajectories,
        id_probabilities,
        {
            "mean": np.nanmean(areas, axis=0),
            "median": np.nanmedian(areas, axis=0),
            "std": np.nanstd(areas, axis=0),
        },
    )


def produce_output_dict(
    blobs_in_video: list[list[Blob]],
    session: Session,
    fragments: list[Fragment] | None = None,
    progress_bar=None,
    abort: Callable = lambda: False,
):
    """Outputs the dictionary with keys: trajectories, git_commit, video_path,
    frames_per_second

    Parameters
    ----------
    blobs_in_video : list
        List of all blob objects (see :class:`~blob.Blobs`) generated by
        considering all the blobs segmented from the video
    session : <Session object>
        See :class:`~video.Video`

    Returns
    -------
    dict
        Output dictionary containing trajectories as values

    """

    centroid_trajectories, id_probabilities, area_stats = produce_trajectories(
        blobs_in_video, session.n_animals, progress_bar, abort, fragments
    )

    if centroid_trajectories is None or abort():
        return None

    output_dict = {
        "trajectories": centroid_trajectories / session.resolution_reduction,
        "version": session.version,
        "video_paths": list(map(str, session.video_paths)),
        "frames_per_second": session.frames_per_second,
        "body_length": session.median_body_length_full_resolution,
        "stats": {"estimated_accuracy": session.estimated_accuracy},
        "areas": area_stats,
        "setup_points": session.setup_points,
        "identities_labels": session.identities_labels
        or [str(i + 1) for i in range(session.n_animals)],
        "identities_groups": {
            key: list(value) for key, value in session.identities_groups.items()
        },
        "length_unit": session.length_unit,
    }

    if id_probabilities is not None and np.isfinite(id_probabilities).any():
        output_dict["id_probabilities"] = id_probabilities
        # After the interpolation some identities that were 0 are assigned
        output_dict["stats"]["estimated_accuracy_after_interpolation"] = (
            1 if session.single_animal else np.nanmean(output_dict["id_probabilities"])
        )
        # Centroids with identity
        identified = ~np.isnan(output_dict["trajectories"][..., 0])
        output_dict["stats"]["percentage_identified"] = np.mean(identified)
        # Estimated accuracy of identified blobs

        output_dict["stats"]["estimated_accuracy_identified"] = (
            1
            if session.single_animal
            else np.nanmean(output_dict["id_probabilities"][identified])
        )

    return output_dict



import logging
from importlib import metadata
from typing import Iterator, Protocol

import torch
from torch.backends import mps

from idtrackerai.utils import IdtrackeraiError, conf


class DataLoaderWithLabels(Protocol):
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[tuple[torch.Tensor, torch.Tensor]]: ...


def get_device(user_device: str) -> torch.device:
    """Returns the current available device for PyTorch"""
    logging.debug("Using PyTorch %s", metadata.version("torch"))
    if user_device:
        try:
            device = torch.device(user_device)
        except RuntimeError as exc:
            raise IdtrackeraiError(
                f'Torch device name "{user_device}" not recognized.'
            ) from exc
        else:
            logging.info('Using user stated device "%s": %s', user_device, repr(device))
            return device
    if torch.cuda.is_available():
        device = torch.device("cuda")
        logging.info('Using Cuda backend with "%s"', torch.cuda.get_device_name(device))
        return device
    if mps.is_available():
        logging.info("Using MacOS Metal backend")
        return torch.device("mps")
    logging.warning(
        "[bold red]No graphic device was found available[/], running neural"
        " networks on CPU. This may slow down the training steps.",
        extra={"markup": True},
    )
    return torch.device("cpu")


DEVICE = get_device(conf.DEVICE)



import numpy as np

from idtrackerai import ListOfFragments
from idtrackerai.utils import conf, track


def compute_model_velocity(
    list_of_fragments: ListOfFragments, percentile=None
) -> float:
    """computes the 2 * (percentile) of the distribution of velocities of identified fish.
    params
    -----
    blobs_in_video: list of blob objects
        collection of blobs detected in the video.
    number_of_animals int
    percentile int
    -----
    return
    -----
    float
    2 * np.max(distance_travelled_in_individual_fragments) if percentile is None
    2 * percentile(velocity distribution of identified animals) otherwise
    """
    if percentile is None:
        percentile = conf.VEL_PERCENTILE

    distance_travelled_in_individual_frag: list[np.ndarray] = []
    for fragment in track(
        list_of_fragments.individual_fragments, "Computing velocity model"
    ):
        distance_travelled_in_individual_frag.append(fragment.frame_by_frame_velocity)

    distances = np.concatenate(distance_travelled_in_individual_frag)
    return (
        2 * distances.max()
        if percentile is None
        else 2 * float(np.percentile(distances, percentile, overwrite_input=True))
    )



from pathlib import Path

import torch
from torch.nn import CrossEntropyLoss
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import GlobalFragment
from idtrackerai.utils import conf, load_id_images

from ..network import (
    CNN,
    DEVICE,
    LearnerClassification,
    NetworkParams,
    StopTraining,
    get_dataloader,
    train_loop,
)
from .identity_dataset import split_data_train_and_validation


def pretrain_global_fragment(
    identification_model: CNN,
    network_params: NetworkParams,
    pretraining_global_fragment: GlobalFragment,
    id_images_file_paths: list[Path],
):
    """Performs pretraining on a single global fragments"""

    images, labels = pretraining_global_fragment.get_images_and_labels()

    images = load_id_images(id_images_file_paths, images)

    (
        train_images,
        train_labels,
        train_weights,
        validation_images,
        validation_labels,
    ) = split_data_train_and_validation(
        images, labels, conf.VALIDATION_PROPORTION, network_params.n_classes
    )

    train_loader = get_dataloader(
        "training", train_images, train_labels, conf.BATCH_SIZE_IDCNN, pretraining=True
    )
    val_loader = get_dataloader("validation", validation_images, validation_labels)

    criterion = CrossEntropyLoss(
        weight=torch.tensor(train_weights, dtype=torch.float32)
    ).to(DEVICE)

    identification_model.fully_connected_reinitialization()

    if network_params.optimizer == "Adam":
        optimizer = torch.optim.Adam(
            identification_model.parameters(), **network_params.optim_args
        )
    elif network_params.optimizer == "SGD":
        optimizer = torch.optim.SGD(
            identification_model.parameters(), **network_params.optim_args
        )
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    learner = LearnerClassification(
        identification_model, criterion, optimizer, scheduler
    )

    stopping = StopTraining(
        epochs_limit=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
        overfitting_limit=conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN,
        plateau_limit=conf.LEARNING_RATIO_DIFFERENCE_IDCNN,
    )

    train_loop(learner, train_loader, val_loader, stopping)
    learner.save_model(network_params.model_path)

    for fragment in pretraining_global_fragment:
        fragment.used_for_pretraining = True



import gc
import logging
from itertools import pairwise

from idtrackerai import ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Session
from idtrackerai.utils import IdtrackeraiError, track


def tracker_API(
    session: Session,
    list_of_blobs: ListOfBlobs,
    list_of_fragments: ListOfFragments,
    list_of_global_fragments: ListOfGlobalFragments,
) -> ListOfFragments:

    if session.track_wo_identities:
        track_without_identities(session, list_of_blobs)

    elif session.single_animal:
        track_single_animal(list_of_blobs)

    elif list_of_global_fragments.no_global_fragment:
        raise IdtrackeraiError(
            "There are no Global Fragments long enough to be candidates"
            " for accumulation, thus it is not possible to train the"
            " identification networks. The video has to contain longer"
            " slices where all animals are visible without crossings."
        )

    elif list_of_global_fragments.single_global_fragment:
        track_single_global_fragment_video(
            session, list_of_blobs, list_of_fragments, list_of_global_fragments
        )

    else:
        from .tracker import TrackerAPI

        logging.info(
            "Deleting ListOfBlobs to save memory, it will be reloaded from disk after"
            " tracking"
        )
        for blob in list_of_blobs.all_blobs:
            blob.__dict__.clear()
        list_of_blobs.blobs_in_video.clear()
        # Blobs contain circular references between them, so the automatic garbage
        # collector won't delete them immediately after the clear().
        # Manually calling gc.collect() is the way to really free RAM
        gc.collect()
        list_of_fragments = TrackerAPI(
            session, list_of_fragments, list_of_global_fragments
        ).track()
        list_of_fragments.update_id_images_dataset()
        gc.collect()  # just in case
        list_of_blobs.blobs_in_video = ListOfBlobs.load(
            session.blobs_path
        ).blobs_in_video

    return list_of_fragments


def track_single_global_fragment_video(
    session: Session,
    list_of_blobs: ListOfBlobs,
    list_of_fragments: ListOfFragments,
    list_of_global_fragments: ListOfGlobalFragments,
):
    logging.info("Tracking single global fragment")
    assert len(list_of_global_fragments) == 1
    global_fragment = list_of_global_fragments.global_fragments[0]

    for identity, fragment in enumerate(global_fragment):
        fragment.temporary_id = identity
        fragment.identity = identity + 1

    session.identities_groups = list_of_fragments.build_exclusive_rois()
    list_of_fragments.update_blobs(list_of_blobs.all_blobs)


def track_single_animal(list_of_blobs: ListOfBlobs):
    logging.info("Tracking a single animal, assigning identity 1 to all blobs")
    for blob in list_of_blobs.all_blobs:
        blob.identity = 1


def track_without_identities(session: Session, list_of_blobs: ListOfBlobs):
    logging.info("Tracking without identities")
    session.number_of_animals = list_of_blobs.max_number_of_blobs_in_one_frame

    current_fragments = [-10 for _ in range(session.number_of_animals)]

    for blobs_in_frame, blobs_in_future in pairwise(
        track(list_of_blobs.blobs_in_video, "Assigning random identities")
    ):
        next_fragments = {b.fragment_identifier for b in blobs_in_future}

        for blob in blobs_in_frame:
            if blob.is_a_crossing:
                continue
            try:
                identity = current_fragments.index(blob.fragment_identifier)
            except ValueError:  # blob's fragment is not in current_fragments
                identity = current_fragments.index(-10)  # look for an empty spot
                current_fragments[identity] = blob.fragment_identifier

            blob.identity = identity + 1
            if blob.fragment_identifier not in next_fragments:
                current_fragments[identity] = -10  # leave an empty spot

    for blob in list_of_blobs.blobs_in_video[-1]:  # last frame
        if blob.is_a_crossing:
            continue
        try:
            identity = current_fragments.index(blob.fragment_identifier)
        except ValueError:  # blob.fragment_identifier is not in identifiers_prev
            identity = current_fragments.index(-10)  # look for an empty spot
            current_fragments[identity] = blob.fragment_identifier
        blob.identity = identity + 1



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
from typing import Sequence

from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import QEvent, QPoint, Qt
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import QHBoxLayout, QSizePolicy, QSlider, QSpinBox, QWidget
from superqt import QLabeledRangeSlider
from superqt.sliders._labeled import LabelPosition


class LabelRangeSlider(QLabeledRangeSlider):
    def __init__(
        self,
        min: int,
        max: int,
        parent: QWidget | None = None,
        start_end_val: tuple[int, int] | None = None,
        block_upper=True,
    ):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.block_upper = block_upper
        self.setRange(min, max)
        self.setValue(start_end_val or (min, max))
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

        self._min_label.setReadOnly(True)
        if block_upper:
            self._max_label.setReadOnly(True)
        else:
            self._max_label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)

        self._handle_labels[0].valueChanged.connect(
            lambda val: self._slider.setSliderPosition(int(val), 0)
        )
        self._handle_labels[1].valueChanged.connect(
            lambda val: self._slider.setSliderPosition(int(val), 1)
        )

        for handle in self._handle_labels:
            handle.setFocusPolicy(Qt.FocusPolicy.WheelFocus)

    def _reposition_labels(self):
        """Overriding superqt method to remove the last label.clearFocus() call"""
        if (
            not self._handle_labels
            or self._handle_label_position == LabelPosition.NoLabel
        ):
            return

        horizontal = self.orientation() == Qt.Orientation.Horizontal
        labels_above = self._handle_label_position == LabelPosition.LabelsAbove

        last_edge = None
        for i, label in enumerate(self._handle_labels):
            rect = self._slider._handleRect(i)
            dx = -label.width() / 2
            dy = -label.height() / 2
            if labels_above:
                if horizontal:
                    dy *= 3
                else:
                    dx *= -1
            else:
                if horizontal:
                    dy *= -1
                else:
                    dx *= 3
            pos = self._slider.mapToParent(rect.center())
            pos += QPoint(int(dx + self.label_shift_x), int(dy + self.label_shift_y))
            if last_edge is not None:
                # prevent label overlap
                if horizontal:
                    pos.setX(int(max(pos.x(), last_edge.x() + label.width() / 2 + 12)))
                else:
                    pos.setY(int(min(pos.y(), last_edge.y() - label.height() / 2 - 4)))
            label.move(pos)
            last_edge = pos
            # label.clearFocus() # better focus behavior without this
            label.show()
        self.update()

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            style = (
                "QDoubleSpinBox{"
                + f"color: #{self.palette().color(QPalette.ColorRole.Text).rgba():x}"
                ";background:transparent; border: 0;"
                f" font-size:{self.font().pointSize()}pt"
                "}QDoubleSpinBox:!enabled{color: #"
                + f"{self.palette().color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).rgba():x}"
                ";}"
            )
            self._slider.setPalette(self.palette())
            self._min_label.setStyleSheet(style)
            self._max_label.setStyleSheet(style)
            self._max_label._update_size()
            self._min_label._update_size()
            for handle in self._handle_labels:
                handle.setStyleSheet(style)
                handle._update_size()

    def value(self) -> tuple[int, int]:
        return super().value()  # type: ignore

    def setValue(self, value: Sequence[int]) -> None:
        if not self.block_upper and value[1] > self.maximum():
            self.setMaximum(value[1])
        return super().setValue(value)  # type: ignore


class InvertibleSlider(QWidget):
    "A labeled slider with the capacity to invert the colored bar without negative numbers"

    valueChanged = Signal(int)

    def __init__(self, min: int, max: int) -> None:
        super().__init__()
        main_layout = QHBoxLayout()
        self.min = min
        self.max = max

        self.slider = QSlider()
        self.slider.setOrientation(Qt.Orientation.Horizontal)
        self.slider.setRange(min, max)

        self.label = QSpinBox()
        self.label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
        self.label.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)
        self.label.setStyleSheet("background:transparent; border: 0;")
        self.label.setRange(min, max)

        main_layout.addWidget(self.slider)
        main_layout.addWidget(self.label)
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.inverted = False

        self.slider.valueChanged.connect(self.slider_changed)
        self.label.valueChanged.connect(self.label_changed)

    def slider_changed(self, value: int):
        self.label.setValue(-value if self.inverted else value)

    def label_changed(self, value: int):
        self.slider.setValue(-value if self.inverted else value)
        self.valueChanged.emit(value)

    def value(self):
        return self.label.value()

    def setValue(self, val: int):
        return self.label.setValue(val)

    def set_inverted(self, inverted: bool):
        """This function must be followed by a valueChanged signal
        I don't implement it here to avoid duplicates and simplify code downstream"""
        self.slider.setInvertedAppearance(inverted)
        self.slider.setInvertedControls(inverted)
        current_value = self.slider.value()
        self.inverted = inverted

        self.slider.blockSignals(True)
        if inverted:
            self.slider.setRange(-self.max, self.min)
            self.slider.setValue(-abs(current_value))
        else:
            self.slider.setRange(self.min, self.max)
            self.slider.setValue(abs(current_value))
        self.slider.blockSignals(False)

    def set_value(self, value: int):
        self.label.setValue(value)



import logging
from contextlib import suppress
from typing import Sequence

from torch import Tensor, nn

from .network_params import NetworkParams


class CNN(nn.Module):
    def __init__(self, input_shape: Sequence[int], out_dim: int):
        logging.info("Creating CNN model")
        super().__init__()

        self.layers = nn.Sequential(
            nn.Conv2d(input_shape[-1], 16, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(16, 64, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(64, 100, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.Flatten(),
            nn.Linear(100 * (input_shape[1] // 4) ** 2, 100),
            nn.ReLU(inplace=True),
            nn.Linear(100, out_dim),
        )

        self.reinitilaize()

    @classmethod
    def from_network_params(cls, network_params: NetworkParams):
        return cls(
            input_shape=network_params.image_size, out_dim=network_params.n_classes
        )

    def forward(self, x: Tensor) -> Tensor:
        # per image normalization
        x -= x.mean((1, 2, 3), keepdim=True)
        with suppress(ValueError):
            x /= x.std((1, 2, 3), keepdim=True)

        return self.layers(x)

    def reinitilaize(self):
        logging.info("Reinitializing model")

        def init_func(m):
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.xavier_uniform_(m.weight.data)

        self.apply(init_func)

    def fully_connected_reinitialization(self):
        logging.info("Reinitializing only fully connected layers")

        def init_func(m):
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)

        self.apply(init_func)



import logging
from typing import Iterable

import numpy as np

from idtrackerai import Blob, ListOfBlobs, ListOfFragments, Session
from idtrackerai.utils import create_dir

from .assign_them_all import close_trajectories_gaps
from .compute_velocity_model import compute_model_velocity
from .correct_impossible_jumps import correct_impossible_velocity_jumps
from .get_trajectories import produce_output_dict
from .trajectories_to_csv import convert_trajectories_file_to_csv_and_json


def trajectories_API(
    session: Session,
    list_of_blobs: ListOfBlobs,
    single_global_fragment: bool,
    list_of_fragments: ListOfFragments,
):
    if (
        not session.track_wo_identities
        and not session.single_animal
        and not single_global_fragment
    ):
        with session.new_timer("Impossible jumps correction"):
            postprocess_impossible_jumps(
                session, list_of_fragments, list_of_blobs.all_blobs
            )

    create_dir(session.trajectories_folder, remove_existing=True)

    trajectories = produce_output_dict(
        list_of_blobs.blobs_in_video, session, list_of_fragments.fragments
    )

    trajectories_file = session.trajectories_folder / "with_gaps.npy"
    logging.info(f"Saving trajectories with gaps in {trajectories_file}")
    np.save(trajectories_file, trajectories)  # type: ignore
    if session.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_file, session.add_time_column_to_csv
        )

    if (
        not session.track_wo_identities
        and not session.single_animal
        and not single_global_fragment
    ):
        interpolate_crossings(session, list_of_blobs, list_of_fragments)
    else:
        list_of_blobs.save(session.blobs_path)
        session.estimated_accuracy = 1.0


def postprocess_impossible_jumps(
    session: Session, list_of_fragments: ListOfFragments, all_blobs: Iterable[Blob]
):
    session.velocity_threshold = compute_model_velocity(list_of_fragments)
    correct_impossible_velocity_jumps(session, list_of_fragments)

    session.individual_fragments_stats = list_of_fragments.get_stats()

    session.estimated_accuracy = compute_estimated_accuracy(list_of_fragments)
    list_of_fragments.save(session.fragments_path)
    list_of_fragments.update_blobs(all_blobs)


def compute_estimated_accuracy(list_of_fragments: ListOfFragments) -> float:
    weighted_P2 = 0
    number_of_individual_blobs = 0

    for fragment in list_of_fragments.individual_fragments:
        if fragment.assigned_identities[0] not in (0, None):
            assert fragment.P2_vector is not None
            weighted_P2 += (
                fragment.P2_vector[fragment.assigned_identities[0] - 1]
                * fragment.n_images
            )
        number_of_individual_blobs += fragment.n_images
    return weighted_P2 / number_of_individual_blobs


def interpolate_crossings(
    session: Session, list_of_blobs: ListOfBlobs, list_of_fragments: ListOfFragments
):
    with session.new_timer("Crossings solver"):
        close_trajectories_gaps(session, list_of_blobs, list_of_fragments)

    list_of_blobs.save(session.blobs_path)
    trajectories_wo_gaps_file = session.trajectories_folder / "without_gaps.npy"
    logging.info(
        "Generating trajectories. The trajectories files are stored in "
        f"{trajectories_wo_gaps_file}"
    )
    trajectories_wo_gaps = produce_output_dict(
        list_of_blobs.blobs_in_video, session, list_of_fragments.fragments
    )
    np.save(trajectories_wo_gaps_file, trajectories_wo_gaps)  # type: ignore
    if session.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_wo_gaps_file, session.add_time_column_to_csv
        )

    # reset crossings to save an improved version of with gaps
    for blob in list_of_blobs.all_blobs:
        if (
            blob.identities_corrected_closing_gaps is not None
            and len(blob.identities_corrected_closing_gaps) > 1
        ):
            blob.identities_corrected_closing_gaps = [0]

    trajectories_file = session.trajectories_folder / "with_gaps.npy"
    logging.info("Saving improved trajectories with gaps")
    trajectories = produce_output_dict(
        list_of_blobs.blobs_in_video, session, list_of_fragments.fragments
    )
    np.save(trajectories_file, trajectories)  # type: ignore
    if session.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_file, session.add_time_column_to_csv
        )



import sys
from argparse import ArgumentParser
from pathlib import Path

from qtpy.QtWidgets import QApplication

from idtrackerai.utils import manage_exception, wrap_entrypoint
from idtrackerai.validator.validation_GUI import ValidationGUI


def input_args():
    parser = ArgumentParser()
    parser.add_argument(
        "session_directory", help="Session directory to validate", type=Path, nargs="?"
    )
    return parser.parse_args()


@wrap_entrypoint
def main():
    # this catches exceptions when raised inside Qt
    def excepthook(exc_type, exc_value, exc_tb):
        assert QApplication  # Pylance is happier with this
        QApplication.quit()
        manage_exception(exc_value)

    sys.excepthook = excepthook

    args = input_args()
    app = QApplication(sys.argv)
    window = ValidationGUI(args.session_directory)
    window.show()
    app.exec()



import json
import logging
import sys
from importlib import metadata
from itertools import count, pairwise
from math import sqrt
from os import cpu_count
from pathlib import Path
from statistics import fmean
from typing import Any, Iterable, Literal, Sequence
from warnings import warn

import cv2
import h5py
import numpy as np

from .utils import (
    Episode,
    IdtrackeraiError,
    LengthCalibration,
    Timer,
    assert_all_files_exist,
    assert_knowledge_transfer_is_possible,
    build_ROI_mask_from_list,
    create_dir,
    json_default,
    json_object_hook,
    remove_dir,
    remove_file,
    resolve_path,
    track,
)


class Session:
    """
    A class containing the main features of the session.

    This class includes properties of the video by itself, user defined
    parameters for the tracking, and other properties that are generated
    throughout the tracking process.

    We use this class as a storage of data coming from different processes.
    However, this is bad practice and it will change in the future.
    """

    velocity_threshold: float
    erosion_kernel_size: int
    ratio_accumulated_images: float
    individual_fragments_stats: dict
    percentage_of_accumulated_images: list[float]
    session_folder: Path
    setup_points: dict[str, list[tuple[int, int]]]
    median_body_length: float
    """median of the diagonals of individual blob's bounding boxes"""
    first_frame_first_global_fragment: list
    identities_groups: dict
    """Named groups of identities stored in the validation GUI.
    If `exclusive ROI`, the identities of each region will be saved here"""
    length_calibrations: list[LengthCalibration]
    """List of length calibrations containing two points (in pixels units) and
    the real distance between these two defined by the user in the Validator."""
    episodes: list[Episode]
    """Indicates the starting and ending frames of each video episode.
    Video episodes are used for parallelization of some processes"""

    original_width: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""
    original_height: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""
    frames_per_second: int
    """Video frame rate in frames per second obtained by OpenCV from the
    video file"""
    accumulation_statistics_data: list[dict[str, list]]
    version: str
    """Version of idtracker.ai"""
    number_of_error_frames: int = -1
    """The number of frames with more blobs than animals. Set on animals_detection."""
    estimated_accuracy: float | None = None
    accumulation_trial: int = 0
    identities_labels: list[str] | None = None
    """A list with a name for every identity. Defined and used in validator"""
    background_from_segmentation_gui: np.ndarray | None = None
    """Background set by segmentation app to save when the app closes"""

    video_paths: list[Path] = []
    """List of paths to the different files the video is composed of.
    If the video is a single file, the list will have length 1"""
    number_of_animals: int = 0
    intensity_ths: None | Sequence[float] = None
    area_ths: None | Sequence[float] = None
    # bkg_model: None | np.ndarray = None
    name: str = ""
    output_dir: Path | None | str = None
    tracking_intervals: list | None = None
    resolution_reduction: float = 1.0
    roi_list: list[str] | str | None = None
    use_bkg: bool = False
    knowledge_transfer_folder: None | Path = None
    check_segmentation: bool = False
    identity_transfer: bool = False
    track_wo_identities: bool = False
    frames_per_episode: int = 500
    background_subtraction_stat: Literal["median", "mean", "max", "min"] = "median"
    number_of_frames_for_background: int = 50
    number_of_parallel_workers: int = 0
    data_policy: Literal[
        "trajectories", "validation", "knowledge_transfer", "idmatcher.ai", "all"
    ] = "idmatcher.ai"
    id_image_size: list[int] = []
    """ Shape of the Blob's identification images (width, height, n_channels)"""
    protocol3_action: Literal["ask", "abort", "continue"] = "ask"
    convert_trajectories_to_csv_and_json: bool = True
    add_time_column_to_csv: bool = False
    """Add a time column (in seconds) to csv trajectory filesy"""
    exclusive_rois: bool = False
    """(experimental feature) Treat each separate ROI as closed identities groups"""
    identity_transfer_succeded: bool = False
    "True if the identity transfer has been done successfully"
    bounding_box_images_in_ram: bool = False
    "Keep bounding box images on RAM and until used, never write them on disk"

    def set_parameters(self, reset: bool = False, **parameters) -> set[str]:
        """Sets parameters to self only if they are present in the class annotations.
        The set of non recognized parameters names is returned"""
        if reset:
            self.__dict__.clear()
        non_recognized_parameters: set[str] = set()
        for param, value in parameters.items():
            lower_param = param.lower()
            if lower_param in self.__class__.__annotations__:
                setattr(self, lower_param, value)
            elif lower_param == "session":
                warn(
                    '"session" parameters is deprecated since v5.2.3, please use "name"'
                )
                self.name = value
            else:
                non_recognized_parameters.add(param)
        return non_recognized_parameters

    def prepare_tracking(self) -> None:
        """Initializes the session object, checking all parameters"""
        logging.debug("Initializing Session")
        self.version = metadata.version("idtrackerai")

        if not isinstance(self.video_paths, list):
            video_paths = [self.video_paths]
        else:
            video_paths = self.video_paths
        self.assert_video_paths(video_paths)
        self.video_paths = [resolve_path(path) for path in video_paths]
        logging.info(
            "Setting video paths to:\n    " + "\n    ".join(map(str, self.video_paths))
        )

        if self.area_ths is None:
            raise IdtrackeraiError("Missing area thresholds parameter")

        if self.intensity_ths is None:
            raise IdtrackeraiError("Missing intensity thresholds parameter")

        self.accumulation_statistics_data = [{}]

        if self.knowledge_transfer_folder is not None:
            self.knowledge_transfer_folder = resolve_path(
                self.knowledge_transfer_folder
            )
            if not self.knowledge_transfer_folder.exists():
                raise IdtrackeraiError(
                    f'Knowledge transfer folder "{self.knowledge_transfer_folder}" not'
                    " found"
                )

            if (
                self.knowledge_transfer_folder.is_dir()
                and self.knowledge_transfer_folder.name.startswith("session_")
            ):
                self.knowledge_transfer_folder /= "accumulation_0"
            self.id_image_size = assert_knowledge_transfer_is_possible(
                self.knowledge_transfer_folder, self.n_animals
            )

        (self.original_width, self.original_height, self.frames_per_second) = (
            self.get_info_from_video_paths(self.video_paths)
        )
        (self.number_of_frames, _, self.tracking_intervals, self.episodes) = (
            self.get_processing_episodes(
                self.video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )

        logging.info(
            f"The session has {self.number_of_frames} "
            f"frames ({self.number_of_episodes} episodes)"
        )
        if len(self.episodes) < 10:
            for episode in self.episodes:
                video_name = episode.video_path.name
                logging.info(
                    f"\tEpisode {episode.index}, frames ({episode.local_start} "
                    f"=> {episode.local_end}) of /{video_name}"
                )
        assert self.number_of_episodes > 0

        if self.output_dir is None:
            self.output_dir = self.video_paths[0].parent
        self.output_dir = resolve_path(self.output_dir)

        if not self.name:
            self.name = "&".join(path.stem for path in self.video_paths)
            logging.info('No session name provided, assigning name "%s"', self.name)

            if (self.output_dir / f"session_{self.name}").exists():
                # add a counter in sessions with default name
                for index in count(1):
                    name = self.name + f"_{index}"
                    if not (self.output_dir / f"session_{name}").exists():
                        break
                else:
                    raise RuntimeError
                logging.info(
                    'A session with the assigned name ("%s") already exists, renaming'
                    ' current session to "%s"',
                    self.name,
                    name,
                )
                self.name = name

        self.session_folder = self.output_dir / f"session_{self.name}"

        create_dir(self.session_folder)
        create_dir(self.preprocessing_folder)

        self.ROI_mask = build_ROI_mask_from_list(
            self.roi_list,
            self.resolution_reduction,
            self.original_width,
            self.original_height,
        )

        if isinstance(self.id_image_size, int):
            self.id_image_size = [self.id_image_size, self.id_image_size, 1]
        elif not self.id_image_size:  # if it is None or empty tuple or list...
            self.id_image_size = []

        if self.number_of_parallel_workers <= 0:
            computer_CPUs = cpu_count()
            if computer_CPUs is not None:
                if self.number_of_parallel_workers == 0:
                    self.number_of_parallel_workers = min((computer_CPUs + 1) // 2, 8)
                elif self.number_of_parallel_workers < 0:
                    self.number_of_parallel_workers += computer_CPUs
        logging.info("Number of parallel jobs: %d", self.number_of_parallel_workers)

        if self.number_of_animals == 0 and not self.track_wo_identities:
            raise IdtrackeraiError(
                "Cannot track with an undefined number of animals (n_animals = 0)"
                " when tracking with identities"
            )

        self.bkg_model = self.background_from_segmentation_gui  # has a setter
        self.__dict__.pop("background_from_segmentation_gui", None)

        self.first_frame_first_global_fragment = []
        self.identities_groups = {}
        self.setup_points = {}

        # Processes timers
        self.timers: dict[str, Timer] = {}

    def new_timer(self, name: str) -> Timer:
        """Generates, saves and returns a Timer"""
        timer = Timer(name)
        self.timers[name] = timer
        return timer

    def __str__(self) -> str:
        return f"<session {self.session_folder}>"

    def set_id_image_size(self, median_body_length: float, reset=False):
        self.median_body_length = median_body_length
        if reset or not self.id_image_size:
            side_length = int(median_body_length / sqrt(2))
            side_length += side_length % 2
            self.id_image_size = [side_length, side_length, 1]
        logging.info(f"Identification image size set to {self.id_image_size}")

    @property
    def n_animals(self) -> int:
        return self.number_of_animals

    @property
    def single_animal(self) -> bool:
        return self.n_animals == 1

    @property
    def bkg_model(self) -> np.ndarray | None:
        if self.background_path.is_file():
            return cv2.imread(str(self.background_path))[..., 0]
        return None

    @bkg_model.setter
    def bkg_model(self, bkg: np.ndarray | None) -> None:
        if bkg is None:
            del self.bkg_model
            return
        # cv2.imwrite has given issues with paths containing chinese characters
        cv2.imencode(self.background_path.suffix, bkg)[1].tofile(self.background_path)
        logging.info(f"Background saved at {self.background_path}")

    @bkg_model.deleter
    def bkg_model(self) -> None:
        self.background_path.unlink(missing_ok=True)

    @property
    def ROI_list(self) -> list[str] | str | None:
        """Fixes compatibility issues"""
        return self.roi_list

    @property
    def ROI_mask(self) -> np.ndarray | None:
        if self.ROI_mask_path.is_file():
            return cv2.imread(str(self.ROI_mask_path))[..., 0]
        return None

    @ROI_mask.setter
    def ROI_mask(self, mask: np.ndarray | None) -> None:
        if mask is None:
            del self.ROI_mask
            return
        # cv2.imwrite has given issues with paths containing chinese characters
        cv2.imencode(self.ROI_mask_path.suffix, mask)[1].tofile(self.ROI_mask_path)
        logging.info(f"ROI mask saved at {self.ROI_mask_path}")

    @ROI_mask.deleter
    def ROI_mask(self) -> None:
        self.ROI_mask_path.unlink(missing_ok=True)

    @property
    def number_of_episodes(self) -> int:
        "Number of episodes in which the video is splitted for parallel processing"
        return len(self.episodes)

    @property
    def width(self) -> int:
        "Video width in pixels after applying the resolution reduction factor"
        return int(self.original_width * self.resolution_reduction + 0.5)

    @property
    def height(self) -> int:
        "Video height in pixels after applying the resolution reduction factor"
        return int(self.original_height * self.resolution_reduction + 0.5)

    @property
    def session(self) -> str:
        warn('"Session.session" is deprecated, please use "Session.name"')
        return self.name

    @property
    def median_body_length_full_resolution(self) -> float:
        """Median body length in pixels in full frame resolution
        (i.e. without considering the resolution reduction factor)
        """
        return self.median_body_length / self.resolution_reduction

    # Paths and folders
    @property
    def preprocessing_folder(self) -> Path:
        return self.session_folder / "preprocessing"

    @property
    def background_path(self) -> Path:
        return self.preprocessing_folder / "background.png"

    @property
    def ROI_mask_path(self) -> Path:
        return self.preprocessing_folder / "ROI_mask.png"

    @property
    def trajectories_folder(self) -> Path:
        return self.session_folder / "trajectories"

    @property
    def crossings_detector_folder(self) -> Path:
        return self.session_folder / "crossings_detector"

    @property
    def pretraining_folder(self) -> Path:
        return self.session_folder / "pretraining"

    @property
    def individual_videos_folder(self) -> Path:
        return self.session_folder / "individual_videos"

    @property
    def accumulation_folder(self) -> Path:
        return self.session_folder / f"accumulation_{self.accumulation_trial}"

    @property
    def id_images_folder(self) -> Path:
        return self.session_folder / "identification_images"

    @property
    def blobs_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs.pickle"

    @property
    def blobs_no_gaps_path(self) -> Path:
        """DEPRECATED since v5.2.2
        get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs_no_gaps.pickle"

    @property
    def blobs_path_validated(self) -> Path:
        "DEPRECATED since v5.2.5. Validated list of blobs are saved in self.blobs_path"
        return self.preprocessing_folder / "list_of_blobs_validated.pickle"

    @property
    def idmatcher_results_path(self) -> Path:
        return self.session_folder / "matching_results"

    @property
    def global_fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_global_fragments.json"

    @property
    def fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_fragments.json"

    @property
    def path_to_session(self) -> Path:
        return self.session_folder / "session.json"

    @property
    def bbox_images_folder(self) -> Path:
        return self.session_folder / "bounding_box_images"

    @property
    def id_images_file_paths(self) -> list[Path]:
        try:
            return [
                self.id_images_folder / f"id_images_{e}.hdf5"
                for e in range(self.number_of_episodes)
            ]
        except AttributeError:
            # Loading a Session without the video files present generates a session
            # without episodes. In this case, lets take all present files in id_images_folder
            paths: list[Path] = []
            for episode in count():
                path = self.id_images_folder / f"id_images_{episode}.hdf5"
                if not path.exists():
                    return paths
                paths.append(path)
            else:
                raise  # for PyLance

    @classmethod
    def defaults(cls) -> dict[str, Any]:
        return {
            key: value
            for key, value in vars(cls).items()
            if not key.startswith("__")
            and not callable(value)
            and not callable(getattr(value, "__get__", None))
        }

    def save(self) -> None:
        """Saves the instantiated Session object"""
        logging.info(f"Saving Session object in {self.path_to_session}", stacklevel=2)
        dict_to_save = (self.defaults() | vars(self)).copy()
        dict_to_save.pop("episodes", None)
        dict_to_save.pop("output_dir", None)
        dict_to_save.pop("background_from_segmentation_gui", None)
        self.path_to_session.write_text(
            json.dumps(dict_to_save, default=json_default, indent=4)
        )

    @classmethod
    def load(cls, path: Path | str, video_paths_dir: Path | None = None) -> "Session":
        """Load a session object stored in a JSON file"""
        path = resolve_path(path)
        logging.info(f"Loading Session from {path}", stacklevel=2)
        if not path.exists():
            raise FileNotFoundError(f"{path} not found")
        if not path.is_file():
            possible_files = ("session.json", "video_object.json", "video_object.npy")
            for file in possible_files:
                if (path / file).is_file():
                    break
            else:
                raise FileNotFoundError(
                    f"Session parameters not fount in folder {path}"
                )
            path /= file

        if path.suffix == ".npy":
            session_dict: dict[str, Any] = cls.open_from_v4(path)
        else:
            with open(path, "r", encoding="utf_8") as file:
                session_dict: dict[str, Any] = json.load(
                    file, object_hook=json_object_hook
                )

        if "n_animals" not in session_dict and "number_of_animals" in session_dict:
            session_dict["n_animals"] = session_dict["number_of_animals"]

        session_dict["video_paths"] = list(
            map(resolve_path, session_dict["video_paths"])
        )

        if "session" in session_dict and "name" not in session_dict:
            # backward compatibility
            session_dict["name"] = session_dict.pop("session")

        session_dict["timers"] = {
            name: Timer.from_dict(timer_dict)
            for name, timer_dict in session_dict.get("timers", {}).items()
        }

        # format timers and Paths
        for key, value in session_dict.items():
            if key.endswith("_timer") and isinstance(value, dict):
                # <=5.2.11 compatibility
                session_dict[key] = Timer.from_dict(value)
            if key.endswith("_folder") and isinstance(value, str):
                session_dict[key] = resolve_path(value)

        if "general_timer" in session_dict:
            # This is the only timer we currently use after tracking,
            # so we want to recover it if it was saved in a previous version style
            session_dict["timers"]["Tracking session"] = session_dict["general_timer"]

        if session_dict.get("length_calibrations"):
            session_dict["length_calibrations"] = [
                LengthCalibration.from_dict(value)
                for value in session_dict["length_calibrations"]
            ]

        session = cls.__new__(cls)
        session.__dict__.update(session_dict)
        session.update_paths(path.parent, video_paths_dir)
        try:
            _, _, _, session.episodes = session.get_processing_episodes(
                session.video_paths,
                session.frames_per_episode,
                session.tracking_intervals,
            )
        except AttributeError:
            logging.warning(
                "Could not load video episodes probably due to loading an old version"
                " session"
            )
        except (IdtrackeraiError, FileNotFoundError) as exc:
            logging.warning("Could not load video episodes. %s", str(exc))

        return session

    @classmethod
    def open_from_v4(cls, path: Path) -> dict:
        from idtrackerai.base import network

        logging.warning("Loading from v4: %s", path)

        # v4 compatibility
        sys.modules["idtrackerai.tracker.network.network_params"] = network
        _dict: dict = np.load(path, allow_pickle=True).item().__dict__
        del sys.modules["idtrackerai.tracker.network.network_params"]

        _dict["version"] = "4.0.12 or below"
        _dict["video_paths"] = list(map(Path, _dict.pop("_video_paths")))
        _dict["session_folder"] = path.parent
        _dict["median_body_length"] = _dict.pop("_median_body_length")
        _dict["frames_per_second"] = _dict.pop("_frames_per_second")
        _dict["original_width"] = _dict.pop("_original_width")
        _dict["original_height"] = _dict.pop("_original_height")
        _dict["number_of_frames"] = _dict.pop("_number_of_frames")
        _dict["identities_groups"] = _dict.pop("_identities_groups")
        _dict["id_image_size"] = list(_dict.pop("_identification_image_size"))
        _dict["setup_points"] = _dict.pop("_setup_points")
        _dict["number_of_animals"] = _dict["_user_defined_parameters"][
            "number_of_animals"
        ]
        _dict["tracking_intervals"] = _dict["_user_defined_parameters"][
            "tracking_interval"
        ]
        _dict["resolution_reduction"] = _dict["_user_defined_parameters"][
            "resolution_reduction"
        ]
        _dict["track_wo_identities"] = _dict["_user_defined_parameters"][
            "track_wo_identification"
        ]
        _dict["accumulation_folder"] = (
            path.parent / Path(_dict.pop("_accumulation_folder")).name
        )
        _dict["_user_defined_parameters"].pop("mask")
        return _dict

    def update_paths(
        self, new_session_path: Path, user_video_paths_dir: Path | None = None
    ) -> None:
        """Update paths of objects (e.g. blobs_path, preprocessing_folder...)
        according to the location of the new Session object given
        by `new_session_path`.
        """
        logging.info(
            f"Searching video files: {[str(path.name) for path in self.video_paths]}"
        )
        folder_candidates: set[Path | None] = {
            user_video_paths_dir,
            self.video_paths[0],
            new_session_path,
            new_session_path.parent,
            self.session_folder.parent,
            self.session_folder,
            Path.cwd(),
        }

        for folder_candidate in folder_candidates:
            if folder_candidate is None:
                continue
            if folder_candidate.is_file():
                folder_candidate = folder_candidate.parent

            candidate_new_video_paths = [
                folder_candidate / path.name for path in self.video_paths
            ]

            try:
                assert_all_files_exist(candidate_new_video_paths)
            except FileNotFoundError:
                continue

            logging.info("All video files found in %s", folder_candidate)
            found = True
            break
        else:
            found = False
            candidate_new_video_paths = []
            logging.error("Video file paths not found: %s", self.video_paths)

        need_to_save = False
        if self.session_folder != new_session_path:
            logging.info(
                f"Updated session folder from {self.session_folder} to"
                f" {new_session_path}"
            )
            self.session_folder = new_session_path
            need_to_save = True

        if found and self.video_paths != candidate_new_video_paths:
            logging.info("Updating new video files paths")
            self.video_paths = candidate_new_video_paths
            need_to_save = True

        if need_to_save:
            self.save()

    @staticmethod
    def assert_video_paths(video_paths: Iterable[Path | str]) -> None:
        if not video_paths:
            raise IdtrackeraiError("Empty Video paths list")

        for path in video_paths:
            path = resolve_path(path)
            if not path.is_file():
                raise IdtrackeraiError(f'Video file "{path}" not found')

            readable = cv2.VideoCapture(str(path)).grab()
            if not readable:
                raise IdtrackeraiError(f'Video file "{path}" not readable by OpenCV.')

    @staticmethod
    def get_info_from_video_paths(
        video_paths: Iterable[Path | str],
    ) -> tuple[int, int, int]:
        """Gets some information about the video from the video file itself.

        Returns:
            width: int, height: int, fps: int
        """

        widths, heights, fps = [], [], []
        for path in video_paths:
            cap = cv2.VideoCapture(str(path))
            widths.append(int(cap.get(3)))
            heights.append(int(cap.get(4)))

            try:
                fps.append(int(cap.get(5)))
            except cv2.error:
                logging.warning(f"Cannot read frame per second for {path}")
                fps.append(None)
            cap.release()

        if len(set(widths)) != 1 or len(set(heights)) != 1:
            raise IdtrackeraiError("Video paths have different resolutions")

        if len(set(fps)) != 1:
            fps = [int(np.mean(fps))]
            logging.warning(
                f"Different frame rates detected ({fps}). "
                f"Setting the frame rate to the mean value: {fps[0]} fps"
            )

        return widths[0], heights[0], fps[0]

    @staticmethod
    def get_processing_episodes(
        video_paths: Sequence[Path | str],
        frames_per_episode: int,
        tracking_intervals=None,
    ) -> tuple[(int, list[int], list[list[int]], list[Episode])]:
        """Process the episodes by getting the number of frames in each video
        path and the tracking interval.

        Episodes are used to compute processes in parallel for different
        parts of the video. They are a tuple with
            (local start frame,
            local end frame,
            video path index,
            global start frame,
            global end frame)
        where "local" means in the specific video path and "global" means in
        the whole (multi path) video

        Episodes are guaranteed to belong to a single video path and to have
        all of their frames (end not included) inside a the tracking interval
        """

        def in_which_interval(frame_number, intervals) -> int | None:
            for i, (start, end) in enumerate(intervals):
                if start <= frame_number < end:
                    return i
            return None

        for path in video_paths:
            if not Path(path).exists():
                raise FileNotFoundError(f"{path} not found")

        # total number of frames for every video path
        video_paths_n_frames = [
            int(cv2.VideoCapture(str(path)).get(7)) for path in video_paths
        ]

        for n_frames, video_path in zip(video_paths_n_frames, video_paths):
            if n_frames <= 0:
                raise IdtrackeraiError(
                    f"OpenCV cannot read the number of frames in {video_path}"
                )
        number_of_frames = sum(video_paths_n_frames)

        # set full tracking interval if not defined
        if tracking_intervals is None:
            tracking_intervals = [[0, number_of_frames]]
        elif isinstance(tracking_intervals[0], int):
            tracking_intervals = [tracking_intervals]

        # find the global frames where the video path changes
        video_paths_changes = [0] + list(np.cumsum(video_paths_n_frames))

        # build an interval list like ("frame" refers to "global frame")
        #   [[first frame of video path 0, last frame of video path 0],
        #    [first frame of video path 1, last frame of video path 1],
        #    [...]]
        video_paths_intervals = list(pairwise(video_paths_changes))

        # find the frames where a tracking interval starts or ends
        tracking_intervals_changes = list(np.asarray(tracking_intervals).flatten())

        # Take into account tracking interval changes
        # and video path changes to compute episodes
        limits = video_paths_changes + tracking_intervals_changes

        # clean repeated limits and sort them
        limits = sorted(set(limits))

        # Create "long episodes" as the intervals between any video path
        # change or tracking interval change (keeping only the ones that
        # are inside a tracking interval)
        long_episodes = []
        for start, end in pairwise(limits):
            if (
                in_which_interval(start, tracking_intervals) is not None
            ) and 0 <= start < number_of_frames:
                long_episodes.append((start, end))

        # build definitive episodes by dividing long episodes to fit in
        # the FRAMES_PER_EPISODE restriction
        index = 0
        episodes = []
        for start, end in long_episodes:
            video_path_index = in_which_interval(start, video_paths_intervals)
            assert video_path_index is not None
            global_local_offset = video_paths_intervals[video_path_index][0]

            n_subepisodes = int((end - start) / (frames_per_episode + 1))
            new_episode_limits = np.linspace(start, end, n_subepisodes + 2, dtype=int)
            for new_start, new_end in pairwise(new_episode_limits):
                episodes.append(
                    Episode(
                        index=index,
                        local_start=new_start - global_local_offset,
                        local_end=new_end - global_local_offset,
                        video_path=resolve_path(video_paths[video_path_index]),
                        global_start=new_start,
                        global_end=new_end,
                    )
                )
                index += 1
        return number_of_frames, video_paths_n_frames, tracking_intervals, episodes

    @staticmethod
    def in_which_interval(frame_number, intervals) -> int | None:
        for i, (start, end) in enumerate(intervals):
            if start <= frame_number < end:
                return i
        return None

    @property
    def length_unit(self) -> float | None:
        """Length calibration factor for translating pixel units to user defined units. Property set in the Validator. Returns None if there are no calibrations."""
        if not hasattr(self, "length_calibrations"):
            return None

        values = []
        for c in self.length_calibrations:
            value = c.value()
            if value is not None:
                values.append(value)
        if not values:
            return None
        return fmean(values)

    def delete_data(self) -> None:
        """Deletes some folders with data, to make the outcome lighter.

        Which folders are deleted depends on the constant DATA_POLICY
        """

        logging.info(f'Data policy: "{self.data_policy}"')

        if self.data_policy == "trajectories":
            remove_dir(self.bbox_images_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
            remove_dir(self.preprocessing_folder)
        elif self.data_policy == "validation":
            remove_dir(self.bbox_images_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
        elif self.data_policy == "knowledge_transfer":
            remove_dir(self.bbox_images_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
        elif self.data_policy == "idmatcher.ai":
            remove_dir(self.bbox_images_folder)
            remove_dir(self.crossings_detector_folder)

    def compress_data(self) -> None:
        """Compress the identification images h5py files"""
        if not self.id_images_folder.exists():
            return

        tmp_path = self.session_folder / "tmp.h5py"

        for path in track(
            self.id_images_file_paths, "Compressing identification images"
        ):
            if not path.is_file():
                continue
            with (
                h5py.File(path, "r") as original_file,
                h5py.File(tmp_path, "w") as compressed_file,
            ):
                for key, data in original_file.items():
                    compressed_file.create_dataset(
                        key, data=data, compression="gzip" if "image" in key else None
                    )
            path.unlink()  # Windows needs this call before rename()
            tmp_path.rename(path)



"""isort:skip_file"""

# NetworkParams should be loaded before LearnerClassification
from torch.backends import cudnn

from .utils import DEVICE, DataLoaderWithLabels
from .network_params import NetworkParams
from .models import CNN
from .learners import LearnerClassification
from .train import (
    train,
    evaluate,
    evaluate_only_acc,
    StopTraining,
    train_loop,
    ImageDataset,
    get_dataloader,
    get_predictions,
    get_onthefly_dataloader,
)

cudnn.benchmark = True  # make it train faster

__all__ = [
    "evaluate",
    "LearnerClassification",
    "train",
    "NetworkParams",
    "DEVICE",
    "CNN",
    "evaluate_only_acc",
    "DataLoaderWithLabels",
    "StopTraining",
    "train_loop",
    "ImageDataset",
    "get_dataloader",
    "get_predictions",
    "get_onthefly_dataloader",
]



import logging
from itertools import pairwise

import cv2
import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QImage, QPainter

from idtrackerai import Session
from idtrackerai.GUI_tools import VideoPathHolder, get_cmap
from idtrackerai.utils import track


def QImageToArray(qimg: QImage) -> np.ndarray:
    width = qimg.width()
    height = qimg.height()
    byte_str = qimg.bits()
    assert byte_str is not None
    return np.frombuffer(byte_str.asstring(height * width * 4), np.uint8).reshape(
        (height, width, 4)
    )[:, :, :-1]


def draw_general_frame(
    np_frame: np.ndarray,
    frame_number: int,
    trajectories: np.ndarray,
    centroid_trace_length: int,
    colors: list[tuple[int, int, int]] | np.ndarray,
    labels: list[str],
) -> np.ndarray:
    ordered_centroid = trajectories[frame_number]
    match np_frame.ndim:
        case 3:
            frame = QImage(
                np_frame.data,
                np_frame.shape[1],
                np_frame.shape[0],
                np_frame.shape[1] * 3,
                QImage.Format.Format_BGR888,
            )
        case 2:
            frame = QImage(
                np_frame.data,
                np_frame.shape[1],
                np_frame.shape[0],
                np_frame.shape[1],
                QImage.Format.Format_Grayscale8,
            )
        case _:
            raise RuntimeError("Invalid frame shape ", np_frame.shape)
    canvas = QImage(frame.size(), QImage.Format.Format_ARGB32_Premultiplied)
    canvas.fill(Qt.GlobalColor.transparent)
    painter = QPainter(canvas)
    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
    pen = painter.pen()
    pen.setWidth(2)
    for cur_id, centroid in enumerate(ordered_centroid):
        if frame_number > centroid_trace_length:
            centroids_trace = trajectories[
                frame_number - centroid_trace_length : frame_number + 1, cur_id
            ]
        else:
            centroids_trace = trajectories[: frame_number + 1, cur_id]
        color = QColor(*colors[cur_id])

        alphas = np.linspace(0, 255, len(centroids_trace), dtype=int)[1:]
        if len(centroids_trace) > 1:
            for alpha, (pointA, pointB) in zip(alphas, pairwise(centroids_trace)):
                if any(pointA < 0) or any(pointB < 0):
                    continue
                color.setAlpha(alpha)
                pen.setColor(color)
                painter.setPen(pen)
                painter.drawLine(*pointA, *pointB)

        if all(centroid > 0):
            color.setAlpha(255)
            painter.setBrush(color)
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawEllipse(centroid[0] - 3, centroid[1] - 3, 6, 6)

    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationOver)
    painter.drawImage(canvas.rect(), frame)
    painter.end()

    arr_img = np.array(QImageToArray(canvas))
    for cur_id, centroid in enumerate(ordered_centroid):
        if all(centroid > 0):
            color = (
                int(colors[cur_id][2]),
                int(colors[cur_id][1]),
                int(colors[cur_id][0]),
            )  # BGR

            arr_img = cv2.putText(
                arr_img,
                labels[cur_id],
                (centroid[0], centroid[1]),
                cv2.FONT_HERSHEY_COMPLEX,
                0.8,
                color,
                2,
            )

    return arr_img


def generate_trajectories_video(
    session: Session,
    trajectories: np.ndarray,
    draw_in_gray: bool,
    centroid_trace_length: int,
    starting_frame: int,
    ending_frame: int,
):
    if draw_in_gray:
        logging.info("Drawing original video in grayscale")

    resize_factor = min(
        1920 / session.original_width, 1080 / session.original_height, 1
    )

    if resize_factor != 1:
        logging.info(f"Applying resize of factor {resize_factor}")

    trajectories = np.nan_to_num(trajectories * resize_factor, nan=-1).astype(int)

    video_name = session.video_paths[0].stem + "_tracked.avi"

    colors = get_cmap(session.n_animals)

    labels = session.identities_labels or list(
        map(str, range(1, session.n_animals + 1))
    )

    path_to_save_video = session.session_folder / video_name

    out_video_width = int(session.original_width * resize_factor + 0.5)
    out_video_height = int(session.original_height * resize_factor + 0.5)

    video_writer = cv2.VideoWriter(
        str(path_to_save_video),
        cv2.VideoWriter.fourcc(*"XVID"),
        session.frames_per_second,
        (out_video_width, out_video_height),
    )

    videoPathHolder = VideoPathHolder(session.video_paths)

    ending_frame = len(trajectories) - 1 if ending_frame is None else ending_frame
    logging.info(f"Drawing from frame {starting_frame} to {ending_frame}")

    for frame in track(range(starting_frame, ending_frame), "Generating video"):
        try:
            img = videoPathHolder.read_frame(frame, not draw_in_gray)
            if resize_factor != 1:
                img = cv2.resize(img, (0, 0), fx=resize_factor, fy=resize_factor)
        except RuntimeError as exc:
            logging.error(str(exc))
            img = np.zeros(
                (
                    (out_video_height, out_video_width)
                    if draw_in_gray
                    else (out_video_height, out_video_width, 3)
                ),
                np.uint8,
            )

        img = draw_general_frame(
            img, frame, trajectories, centroid_trace_length, colors, labels
        )

        video_writer.write(img)

    logging.info(f"Video generated in {path_to_save_video}")



import logging
import pickle
from contextlib import suppress
from io import BytesIO
from itertools import chain, pairwise, product
from multiprocessing import Pool
from pathlib import Path

import h5py
import numpy as np

from . import Blob
from .utils import Episode, clean_attrs, resolve_path, track


class ListOfBlobs:
    """Contains all the instances of the class :class:`~blob.Blob` for all
    frames in the video.
    """

    blobs_are_connected: bool

    blobs_in_video: list[list[Blob]]

    def __init__(self, blobs_in_video: list[list[Blob]]):
        logging.info("Generating ListOfBlobs object")
        self.blobs_in_video = blobs_in_video
        self.blobs_are_connected = False

    @property
    def all_blobs(self):
        return chain.from_iterable(self.blobs_in_video)

    @property
    def number_of_blobs(self) -> int:
        return sum(map(len, self.blobs_in_video))

    @property
    def number_of_crossing_blobs(self) -> int:
        return sum(blob.is_a_crossing for blob in self.all_blobs)

    @property
    def number_of_frames(self):
        return len(self.blobs_in_video)

    @property
    def max_number_of_blobs_in_one_frame(self):
        return max(map(len, self.blobs_in_video))

    def __len__(self):
        return len(self.blobs_in_video)

    def compute_overlapping_between_subsequent_frames(self):
        """Computes overlapping between blobs in consecutive frames.

        Two blobs in consecutive frames overlap if the intersection of the list
        of pixels of both blobs is not empty.

        See Also
        --------
        :meth:`blob.Blob.overlaps_with`
        """
        if self.blobs_are_connected:
            logging.error("List of blobs is already connected")
            return
        # self.disconnect()

        for blobs, blobs_next in pairwise(
            track(self.blobs_in_video, "Connecting blobs")
        ):
            for blob, blob_next in product(blobs, blobs_next):
                if blob.overlaps_with(blob_next):
                    blob.now_points_to(blob_next)
        self.blobs_are_connected = True

        # clean cached property
        with suppress(AttributeError):
            for blob in self.all_blobs:
                del blob.convexHull

    def save(self, path: Path | str):
        """Saves instance of the class

        Parameters
        ----------
        path_to_save : str, optional
            Path where to save the object, by default None
        """
        path = resolve_path(path)
        logging.info(f"Saving ListOfBlobs at {path}", stacklevel=2)
        path.parent.mkdir(exist_ok=True)
        self.disconnect()

        for blob in self.all_blobs:
            clean_attrs(blob)

        with open(path, "wb") as file:
            pickle.dump(self, file, protocol=pickle.HIGHEST_PROTOCOL)
        self.reconnect()

    @classmethod
    def load(cls, path: Path | str) -> "ListOfBlobs":
        """Loads an instance of a class saved in a .npy file.

        Parameters
        ----------
        blob_list_file : Path
            path to a saved instance of a ListOfBlobs object

        Returns
        -------
        ListOfBlobs
        """
        path = resolve_path(path)
        logging.info(f"Loading ListOfBlobs from {path}", stacklevel=2)
        if not path.is_file():
            v4_path = path.with_name(
                path.name.replace("list_of_blobs", "blobs_collection")
            ).with_suffix(".npy")

            if v4_path.is_file():
                list_of_blobs = cls.load_from_v4(v4_path)
            else:
                raise FileNotFoundError(path)
        else:
            with open(path, "rb") as file:
                list_of_blobs: ListOfBlobs = pickle.load(file)
        list_of_blobs.reconnect()
        return list_of_blobs

    @classmethod
    def load_from_v4(cls, path: Path) -> "ListOfBlobs":
        logging.info("Loading from v4 file: %s", path)
        list_of_blobs: "ListOfBlobs" = np.load(path, allow_pickle=True).item()

        for blob in track(
            list_of_blobs.all_blobs, "Updating objects from an old idtracker.ai version"
        ):
            blob.is_an_individual = blob._is_an_individual  # type:ignore
            blob.fragment_identifier = blob._fragment_identifier  # type:ignore
            blob.identity = blob._identity  # type:ignore
            blob.identity_corrected_solving_jumps = (
                blob._identity_corrected_solving_jumps  # type:ignore
            )
            blob.identities_corrected_closing_gaps = (
                blob._identities_corrected_closing_gaps  # type:ignore
            )
        return list_of_blobs

    def disconnect(self):
        if self.blobs_are_connected:
            for blob in self.all_blobs:
                blob.next = ()

    def reconnect(self):
        if isinstance(next(self.all_blobs).next, list):
            logging.info("Converting ListOfBlobs from version older than 5.2.2")
            for blob in self.all_blobs:
                blob.previous = tuple(blob.previous)
                blob.next = ()

        if self.blobs_are_connected:
            for blob in self.all_blobs:
                for prev_blob in blob.previous:
                    prev_blob.next = prev_blob.next + (blob,)

    def set_images_for_identification(
        self,
        episodes: list[Episode],
        id_images_files: list[Path],
        id_image_size: list[int],
        n_jobs: int,
    ) -> None:
        """Computes and saves the images used to classify blobs as crossings
        and individuals and to identify the animals along the video.

        Parameters
        ----------
        episodes_start_end : list
            List of tuples of integers indncating the starting and ending
            frames of each episode.
        id_images_file_paths : list
            List of strings indicating the paths to the files where the
            identification images of each episode are stored.
        id_image_size : tuple
            Tuple indicating the width, height and number of channels of the
            identification images.
        number_of_animals : int
            Number of animals to be tracked as indicated by the user.
        number_of_frames : int
            Number of frames in the video
        video_path : str
            Path to the video file
        height : int
            Height of a video frame considering the resolution reduction
            factor.
        width : int
            Width of a video frame considering the resolution reduction factor.
        """

        inputs = [
            (
                id_image_size[0],
                id_images_file,
                episode,
                self.blobs_in_video[episode.global_start : episode.global_end],
            )
            for id_images_file, episode in zip(id_images_files, episodes)
        ]

        if n_jobs == 1:
            for input in track(inputs, "Setting images for identification"):
                self.set_id_images_per_episode(input)
        else:
            with Pool(n_jobs, maxtasksperchild=3) as p:
                for _ in track(
                    p.imap_unordered(self.set_id_images_per_episode, inputs),
                    "Setting images for identification",
                    len(inputs),
                ):
                    pass

        for input in inputs:
            episode, blobs_in_episode = input[2:]
            if isinstance(episode.bbox_images, BytesIO):
                episode.bbox_images.close()
            for index, blob in enumerate(chain.from_iterable(blobs_in_episode)):
                blob.id_image_index = index
                blob.episode = episode.index

    @staticmethod
    def set_id_images_per_episode(
        inputs: tuple[int, Path, Episode, list[list[Blob]]]
    ) -> None:
        id_image_size, file_path, episode, blobs_in_episode = inputs

        with (
            h5py.File(file_path, "w") as id_images_file,
            h5py.File(episode.bbox_images, "r") as bbox_images_file,
        ):
            imgs_to_save = id_images_file.create_dataset(
                "id_images",
                (sum(map(len, blobs_in_episode)), id_image_size, id_image_size),
                np.uint8,
            )

            for index, blob in enumerate(chain.from_iterable(blobs_in_episode)):
                bbox_image: np.ndarray = bbox_images_file[blob.bbox_img_id][:]  # type: ignore
                imgs_to_save[index] = blob.get_image_for_identification(
                    id_image_size, bbox_image
                )

    # TODO: maybe move to crossing detector
    def update_id_image_dataset_with_crossings(self, id_images_file_paths: list[Path]):
        """Adds a array to the identification images files indicating whether
        each image is an individual or a crossing.
        """
        logging.info("Updating crossings in identification images files")

        crossings = []
        for path in id_images_file_paths:
            with h5py.File(path, "r") as file:
                crossings.append(np.empty(len(file["id_images"]), bool))  # type: ignore

        for blob in self.all_blobs:
            crossings[blob.episode][blob.id_image_index] = blob.is_a_crossing

        for path, crossing in zip(id_images_file_paths, crossings):
            try:
                with h5py.File(path, "r+") as file:
                    file.create_dataset("crossings", data=crossing)
            except BlockingIOError as exc:
                # Some MacOS crash with
                # BlockingIOError: [Errno 35] Unable to open file (unable to lock file, errno = 35, error message = 'Resource temporarily unavailable')
                logging.error(f"Failed at writting in {path}: {exc}")

    def remove_centroid(self, frame_number: int, centroid_to_remove, id_to_remove):
        for blob in self.blobs_in_video[frame_number]:
            for indx, (id, centroid) in enumerate(
                zip(blob.all_final_identities, blob.all_final_centroids)
            ):
                if id == id_to_remove:
                    dist = (centroid[0] - centroid_to_remove[0]) ** 2 + (
                        centroid[1] - centroid_to_remove[1]
                    ) ** 2
                    if dist < 1:  # it is the same centroid
                        blob.init_validator_variables()
                        blob.user_generated_centroids[indx] = (-1, -1)
                        blob.user_generated_identities[indx] = -1

    def reset_user_generated_corrections(
        self, start_frame: int = 0, end_frame: int | None = None
    ):
        """[Validation] Resets the identities and centroids generetad by the user.

        Parameters
        ----------
        start_frame : int
            Frame from which to start resetting identities and centroids
        end_frame : int
            Frame where to end resetting identities and centroids
        """

        for blobs_in_frame in track(
            self.blobs_in_video[start_frame:end_frame], "Resetting user corrections"
        ):
            # Reset all user generated identities and centroids
            for blob in blobs_in_frame:
                if blob.added_by_user:
                    self.blobs_in_video[blob.frame_number].remove(blob)
                else:
                    blob.user_generated_identities = None  # type: ignore
                    blob.user_generated_centroids = None  # type: ignore

    def update_centroid(
        self, frame_number: int, centroid_id: int, old_centroid, new_centroid
    ):
        old_centroid = tuple(old_centroid)
        new_centroid = tuple(new_centroid)
        assert len(old_centroid) == 2
        assert len(new_centroid) == 2
        blobs_in_frame = self.blobs_in_video[frame_number]
        assert blobs_in_frame

        dist_to_old_centroid: list[tuple[Blob, float]] = []

        for blob in blobs_in_frame:
            try:
                indx, centroid, dist = blob.index_and_centroid_closer_to(
                    old_centroid, centroid_id
                )
            except ValueError:  # blob has not centroid_id
                pass
            else:
                dist_to_old_centroid.append((blob, dist))

        blob_with_old_centroid = min(dist_to_old_centroid, key=lambda x: x[1])[0]
        blob_with_old_centroid.update_centroid(old_centroid, new_centroid, centroid_id)

    def add_centroid(self, frame_number: int, identity: int, centroid):
        centroid = tuple(centroid)
        assert len(centroid) == 2
        blobs_in_frame = self.blobs_in_video[frame_number]
        if not blobs_in_frame:
            self.add_blob(frame_number, centroid, identity)
            return

        for blob in blobs_in_frame:
            if blob.contains_point(centroid):
                blob.add_centroid(centroid, identity)
                return

        blob = min(blobs_in_frame, key=lambda b: b.distance_from_countour_to(centroid))
        blob.add_centroid(centroid, identity)

    def add_blob(self, frame_number: int, centroid: tuple, identity: int):
        """[Validation] Adds a Blob object the frame number.

        Adds a Blob object to a given frame_number with a given centroid and
        identity. Note that this Blob won't have most of the features (e.g.
        area, contour, fragment_identifier, bbox, ...). It is only
        intended to be used for validation and correction of trajectories.
        The new blobs generated are considered to be individuals.

        Parameters
        ----------
        frame_number : int
            Frame in which the new blob will be added
        centroid : tuple
            The centroid of the new blob
        identity : int
            Identity of the new blob
        """
        contour = np.array(
            [
                [centroid[0] - 2, centroid[1] - 2],
                [centroid[0] - 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] - 2],
            ],
            int,
        )
        new_blob = Blob(contour, frame_number)
        new_blob.added_by_user = True
        new_blob.user_generated_centroids = [(centroid[0], centroid[1])]
        new_blob.user_generated_identities = [identity]
        new_blob.is_an_individual = True
        self.blobs_in_video[frame_number].append(new_blob)



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
import numpy as np
from qtpy.QtCore import QEvent, QPointF, Qt
from qtpy.QtGui import QKeyEvent, QPainterPath, QPalette, QPolygonF, QResizeEvent
from qtpy.QtWidgets import QDialog, QFrame, QHBoxLayout, QLabel, QSizePolicy, QStyle

from idtrackerai.utils import get_vertices_from_label


class LightPopUp(QDialog):
    """A light version of QMessageBox.warning()"""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowFlags(Qt.WindowType.Popup)
        self.setLayout(QHBoxLayout())
        self.text = QLabel("")
        self.text.setWordWrap(True)
        self.icon = QLabel()

        self.icon.setFixedSize(100, 100)
        self.setFixedWidth(500)

        self.layout().addWidget(self.icon)
        self.layout().addWidget(self.text)

    def warning(self, title: str, text) -> None:
        self.icon.setPixmap(
            self.style()
            .standardIcon(QStyle.StandardPixmap.SP_MessageBoxWarning)
            .pixmap(70, 70)
        )
        self.text.setText(f"<strong><center>{title}</strong></center><br><br>{text}")
        self.exec()

    def info(self, title: str, text) -> None:
        self.icon.setPixmap(
            self.style()
            .standardIcon(QStyle.StandardPixmap.SP_MessageBoxInformation)
            .pixmap(70, 70)
        )
        self.text.setText(f"<strong><center>{title}</strong></center><br><br>{text}")
        self.exec()

    def keyPressEvent(self, *args, **kwargs) -> None:
        self.close()


class QHLine(QFrame):
    def __init__(self) -> None:
        super().__init__()
        self.setFrameShape(QFrame.Shape.HLine)
        self.setContentsMargins(10, 0, 10, 0)
        self.setEnabled(False)

    def changeEvent(self, event: QEvent) -> None:
        if event.type() == QEvent.Type.EnabledChange:
            self.setEnabled(False)


def build_ROI_patches_from_list(
    list_of_ROIs: list[str] | str | None,
    resolution_reduction: float,
    width: int,
    height: int,
) -> QPainterPath:
    path = QPainterPath()

    if list_of_ROIs is None:
        return path

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = [list_of_ROIs]

    path.addRect(
        -0.5, -0.5, width * resolution_reduction, height * resolution_reduction
    )

    for line in list_of_ROIs:
        path_i = get_path_from_points(
            get_vertices_from_label(line), resolution_reduction
        )

        if line[0] == "+":
            path -= path_i
        elif line[0] == "-":
            path += path_i
        else:
            raise TypeError
    return path


def get_path_from_points(points: np.ndarray, res_reduct: float = 1) -> QPainterPath:
    points = points * res_reduct + 0.5

    path = QPainterPath()
    if points.ndim == 2:
        # some polygons are made from a single point, 1 dimension
        path.addPolygon(QPolygonF(QPointF(*point) for point in points))
    return path.simplified()


class WrappedLabel(QLabel):
    def __init__(
        self,
        text: str = "",
        framed: bool = False,
        align: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignLeft,
    ) -> None:
        super().__init__(text)
        if framed:
            self.setBackgroundRole(QPalette.ColorRole.Base)
            self.setAutoFillBackground(True)
            self.setContentsMargins(5, 3, 5, 3)
        self.setAlignment(align)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)

    def set_size(self) -> None:
        self.setMinimumHeight(0)
        self.setMinimumHeight(max(self.heightForWidth(self.width()), 1))

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.set_size()
        super().resizeEvent(a0)

    def setText(self, text: str) -> None:
        # Add Zero-width space in backslashes for proper word wrapping
        super().setText(text.replace("\\", "\\\u200b"))
        self.set_size()

    def text(self):
        return super().text().replace("\u200b", "")


_ignored_keys = {
    Qt.Key.Key_D,
    Qt.Key.Key_A,
    Qt.Key.Key_Left,
    Qt.Key.Key_Right,
    Qt.Key.Key_0,
    Qt.Key.Key_1,
    Qt.Key.Key_2,
    Qt.Key.Key_3,
    Qt.Key.Key_4,
    Qt.Key.Key_5,
    Qt.Key.Key_6,
    Qt.Key.Key_7,
    Qt.Key.Key_8,
    Qt.Key.Key_9,
}


def key_event_modifier(event: QKeyEvent) -> QKeyEvent | None:
    event.key().numerator
    if event.key() == Qt.Key.Key_W:
        return QKeyEvent(event.type(), Qt.Key.Key_Up, event.modifiers())
    if event.key() == Qt.Key.Key_S:
        return QKeyEvent(event.type(), Qt.Key.Key_Down, event.modifiers())
    if event.key() in _ignored_keys:
        # These keys would be accepted by QTableWidget
        # but we want them to control the VideoPlayer
        event.ignore()
        return None
    return event



import logging
import os
from datetime import datetime
from functools import wraps
from importlib import metadata
from pathlib import Path
from platform import platform, python_version
from shutil import copy
from traceback import extract_tb
from typing import Callable

from rich.console import Console, ConsoleRenderable
from rich.logging import RichHandler

from .check_PyPI_version import check_version_on_console_thread
from .py_utils import IdtrackeraiError, resolve_path

LOG_FILE_PATH = resolve_path("idtrackerai.log")

ERROR_MSG = (
    "\n\nIf this error happens right after the installation,"
    " check our installation troubleshooting guide"
    " https://idtracker.ai/latest/install/installation_troubleshooting.html"
    "\n\nIf this error persists please let us know by following any of the following"
    " options:\n  - Posting on https://groups.google.com/g/idtrackerai_users\n  -"
    " Opening an issue at https://gitlab.com/polavieja_lab/idtrackerai\n  - Sending an"
    f" email to idtrackerai@gmail.com\nShare the log file ({LOG_FILE_PATH}) when doing"
    " any of the options above"
)

LEVEL_FORMAT = {
    "WARNING": "red",
    "WARN": "red",
    "ERROR": "red bold",
    "FATAL": "red bold reverse",
    "CRITICAL": "red bold reverse",
    "NOTSET": "red",
}


class LevelRichHandler(RichHandler):
    """Modified Rich Logging Handler that prints the logging level only if its
    equal or grater than a warning. Intended to work with `show_level=False`"""

    def render_message(
        self, record: logging.LogRecord, message: str
    ) -> ConsoleRenderable:
        """Method override to manage log level only if its >= warning

        Args:
            record (LogRecord): logging Record.
            message (str): String containing log message.

        Returns:
            ConsoleRenderable: Renderable to display log message.
        """
        if record.levelno >= logging.WARNING:
            record.markup = True
            format = LEVEL_FORMAT.get(record.levelname, "red")
            message = f"[{format}]{record.levelname}[/{format}] {message}"
        return super().render_message(record, message)


def initLogger(level: int = logging.DEBUG):
    logger_width_when_no_terminal = 126
    try:
        os.get_terminal_size()
    except OSError:
        # stdout is sent to file. We define logger width to a constant
        size = logger_width_when_no_terminal
    else:
        # stdout is sent to terminal
        # We define logger width to adapt to the terminal width
        size = None

    LOG_FILE_PATH.unlink(True)  # avoid conflicts and merged files

    # The first handler is the terminal, the second one the .log file,
    # both rendered with Rich and full logging (level=0)
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="%H:%M:%S",
        force=True,
        handlers=[
            LevelRichHandler(
                console=Console(width=size), rich_tracebacks=True, show_level=False
            ),
            LevelRichHandler(
                console=Console(
                    file=LOG_FILE_PATH.open("w", encoding="utf_8"),  # noqa SIM115
                    width=logger_width_when_no_terminal,
                ),
                show_level=False,
            ),
        ],
    )

    os.environ["OPENCV_FFMPEG_LOGLEVEL"] = "-8"  # avoid huge logs with corrupted videos
    logging.captureWarnings(True)
    logging.info("[bold]Welcome to idtracker.ai", extra={"markup": True})
    logging.debug(
        f"Running idtracker.ai '{metadata.version('idtrackerai')}'"
        f" on Python '{python_version()}'\nPlatform: '{platform(True)}'"
        "\nDate: " + str(datetime.now()).split(".")[0]
    )
    logging.info("Writing log in %s", LOG_FILE_PATH)


def wrap_entrypoint(main_function: Callable):
    @wraps(main_function)
    def ret_fun(*args, **kwargs):
        initLogger()
        check_version_on_console_thread()
        try:
            return main_function(*args, **kwargs)
        except (Exception, KeyboardInterrupt) as exc:
            manage_exception(exc)
            if hasattr(exc, "log_path"):
                copy(LOG_FILE_PATH, exc.log_path)  # type: ignore
                logging.info(f"Log file copied to {exc.log_path}")  # type: ignore
            return False

    return ret_fun


def manage_exception(exc: BaseException):
    match exc:
        case IdtrackeraiError():
            tb = extract_tb(exc.__traceback__)[-1]
            logging.critical(
                "%s [bright_black](from %s:%d)[/]",
                exc,
                Path(*Path(tb.filename).parts[-2:]),
                tb.lineno,
                extra={"markup": True},
            )
        case KeyboardInterrupt():
            logging.critical("KeyboardInterrupt", exc_info=False)
            return
        case ModuleNotFoundError():
            if "torch" in str(exc):
                logging.critical(
                    "Module PyTorch is not installed, follow their guideline to install"
                    " it (https://pytorch.org/get-started/locally/). Original"
                    ' exception: "%s"',
                    exc,
                )
                return
            logging.critical("%s: %s", type(exc).__name__, exc, exc_info=exc)
            logging.info(ERROR_MSG)
            return
        case Exception():
            logging.critical("%s: %s", type(exc).__name__, exc, exc_info=exc)
            logging.info(ERROR_MSG)
            return



from typing import Literal, Sequence

import numpy as np

from . import Blob, Fragment


class GlobalFragment:
    """Represents a collection of :class:`fragment.Fragment` N different
    animals. Where N is the number of animals in the video as defined by the
    user.
    """

    duplicated_identities: set

    first_frame_of_the_core: int

    fragments_identifiers: Sequence[int]

    fragments: list[Fragment]

    minimum_distance_travelled: float

    accumulation_step: int | None = None
    """Integer indicating the accumulation step at which the global fragment
    was globally accumulated."""

    def __init__(
        self,
        blobs_in_video: list[list[Blob]],
        fragments: list[Fragment],
        first_frame_of_the_core: int,
    ):
        self.first_frame_of_the_core = first_frame_of_the_core
        self.fragments_identifiers = tuple(
            blob.fragment_identifier for blob in blobs_in_video[first_frame_of_the_core]
        )
        self.set_individual_fragments(fragments)

        for fragment in self:
            fragment.is_in_a_global_fragment = True

        self.minimum_distance_travelled = min(
            fragment.distance_travelled for fragment in self
        )

    @property
    def min_n_images_per_fragment(self):
        return min(fragment.n_images for fragment in self)

    def __iter__(self):
        return iter(self.fragments)

    @classmethod
    def from_json(cls, data: dict, fragments: list[Fragment] | None):
        global_fragment = cls.__new__(cls)
        if "individual_fragments_identifiers" in data:
            data["fragments_identifiers"] = data.pop("individual_fragments_identifiers")
        global_fragment.__dict__.update(data)
        if "duplicated_identities" in data:
            global_fragment.duplicated_identities = set(data["duplicated_identities"])

        if fragments is not None:
            global_fragment.set_individual_fragments(fragments)

        return global_fragment

    @property
    def used_for_training(self):
        """Boolean indicating if all the fragments in the global fragment
        have been used for training the identification network"""
        return all(fragment.used_for_training for fragment in self)

    def is_unique(self, number_of_animals: int):
        """Boolean indicating that the global fragment has unique
        identities, i.e. it does not have duplications."""
        return {fragment.temporary_id for fragment in self} == set(
            range(number_of_animals)
        )

    @property
    def is_partially_unique(self):
        """Boolean indicating that a subset of the fragments in the global
        fragment have unique identities"""

        identities_acceptable_for_training = [
            fragment.temporary_id
            for fragment in self
            if fragment.acceptable_for_training
        ]
        self.duplicated_identities = {
            x
            for x in identities_acceptable_for_training
            if identities_acceptable_for_training.count(x) > 1
        }
        return len(self.duplicated_identities) == 0

    def set_individual_fragments(self, fragments: Sequence[Fragment]):
        """Gets the list of instances of the class :class:`fragment.Fragment`
        that constitute the global fragment and sets an attribute with such
        list.

        Parameters
        ----------
        fragments : list
            All the fragments extracted from the video.

        """
        self.fragments = [
            fragments[identifier] for identifier in self.fragments_identifiers
        ]

    def acceptable_for_training(
        self, accumulation_strategy: Literal["global", "partial"]
    ) -> bool:
        """Returns True if the global fragment is acceptable for training"""

        return (all if accumulation_strategy == "global" else any)(
            fragment.acceptable_for_training for fragment in self
        )

    @property
    def total_number_of_images(self) -> int:
        """Gets the total number of images in the global fragment"""
        return sum(fragment.n_images for fragment in self)

    def get_images_and_labels(self):
        """Gets the images and identities in the global fragment as a
        labelled dataset in order to train the identification neural network

        If the scope is "pretraining" the identities of each fragment
        will be arbitrary.
        If the scope is "identity_transfer" then the labels will be
        empty as they will be infered by the identification network selected
        by the user to perform the transferring of identities.

        Parameters
        ----------
        id_images_file_paths : list
            List of paths (str) where the identification images are stored.
        scope : str, optional
            Whether the images are going to be used for training the
            identification network or for "pretraining", by default
            "pretraining".

        Returns
        -------
        Tuple
            Tuple with two Numpy arrays with the images and their labels.
        """
        images: list[tuple[int, int]] = []
        labels: list[int] = []

        for temporary_id, fragment in enumerate(self):
            images += fragment.image_locations
            labels += [temporary_id] * fragment.n_images

        return np.asarray(images), np.asarray(labels)



from contextlib import suppress
from importlib import metadata
from warnings import warn

with suppress(ImportError):
    # PyQt has to be imported before CV2 (importing idtrackerai stuff implies CV2)
    # If not, the QFileDialog.getFileNames() does not load the icons, very weird
    from qtpy.QtWidgets import QApplication  # noqa F401

# Video has to be the first class to be imported
from idtrackerai.session import Session

from .blob import Blob
from .fragment import Fragment
from .globalfragment import GlobalFragment
from .list_of_blobs import ListOfBlobs
from .list_of_fragments import ListOfFragments
from .list_of_global_fragments import ListOfGlobalFragments

__version__ = metadata.version("idtrackerai")


class Video(Session):
    "Backward compatibility since the rename of the `Video` class for `Session`"

    def __new__(cls):
        warn("Video is deprecated since v5.2.3, it has been renamed to `Session`")
        return super().__new__(cls)


__all__ = [
    "Blob",
    "ListOfBlobs",
    "ListOfFragments",
    "ListOfGlobalFragments",
    "ListOfGlobalFragments",
    "GlobalFragment",
    "Session",
    "Fragment",
    "Video",
]



import logging

import numpy as np

from idtrackerai import ListOfFragments, ListOfGlobalFragments, Session
from idtrackerai.utils import IdtrackeraiError, conf, create_dir

from ..network import CNN, DEVICE, LearnerClassification, NetworkParams
from .accumulation_manager import AccumulationManager
from .accumulator import perform_one_accumulation_step
from .assigner import assign_remaining_fragments
from .identity_transfer import identify_first_global_fragment_for_accumulation
from .pre_trainer import pretrain_global_fragment


class TrackerAPI:
    "API for tracking with identities more than one animal with more than one Global Fragment"

    identification_model: CNN
    accumulation_network_params: NetworkParams

    def __init__(
        self,
        session: Session,
        list_of_fragments: ListOfFragments,
        list_of_global_fragments: ListOfGlobalFragments,
    ):
        self.session = session
        self.list_of_fragments = list_of_fragments
        self.list_of_global_fragments = list_of_global_fragments

    def track(self) -> ListOfFragments:
        """In protocol 3, list_of_fragments is loaded from accumulation
        folders so the reference from outside tracker_API is lost.
        That's why list_of_fragments has to be returned"""
        logging.info("Tracking with identities")
        self.session.accumulation_trial = 0
        create_dir(self.session.accumulation_folder, remove_existing=True)
        self.accumulation_network_params = NetworkParams(
            n_classes=self.session.n_animals,
            save_folder=self.session.accumulation_folder,
            knowledge_transfer_folder=self.session.knowledge_transfer_folder,
            model_name="identification_network",
            image_size=self.session.id_image_size,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
        )
        self.accumulation_network_params.save()
        with self.session.new_timer("Accumulation"):
            self.accumulation_protocol()
        with self.session.new_timer("Identification"):
            assign_remaining_fragments(
                self.list_of_fragments,
                self.identification_model,
                self.accumulation_network_params,
            )
        return self.list_of_fragments

    def accumulation_protocol(self) -> None:

        self.list_of_fragments.reset(roll_back_to="fragmentation")

        if self.session.knowledge_transfer_folder:
            try:
                self.identification_model = LearnerClassification.load_model(
                    self.accumulation_network_params, knowledge_transfer=True
                )
                logging.info("Tracking with knowledge transfer")
                if not self.session.identity_transfer:
                    self.identification_model.fully_connected_reinitialization()
                else:
                    logging.info(
                        "Identity transfer. Not reinitializing the fully connected"
                        " layers."
                    )
            except RuntimeError as exc:
                logging.error(
                    f"Could not load model {self.accumulation_network_params} to"
                    " transfer knowledge from, following without knowledge nor identity"
                    " transfer.\n"
                    f"Raised error: {exc}"
                )
                self.identification_model = CNN.from_network_params(
                    self.accumulation_network_params
                ).to(DEVICE)
        else:
            self.identification_model = CNN.from_network_params(
                self.accumulation_network_params
            ).to(DEVICE)

        first_global_fragment = max(
            self.list_of_global_fragments, key=lambda gf: gf.minimum_distance_travelled
        )

        self.session.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
        )

        identify_first_global_fragment_for_accumulation(
            first_global_fragment,
            self.session,
            identification_model=self.identification_model,
        )

        self.session.identities_groups = self.list_of_fragments.build_exclusive_rois()

        # Order global fragments by distance to the first global fragment for the accumulation
        self.list_of_global_fragments.sort_by_distance_to_the_frame(
            first_global_fragment.first_frame_of_the_core
        )

        # Instantiate accumulation manager
        self.accumulation_manager = AccumulationManager(
            self.session.n_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        # Selecting the first global fragment is considered as
        # the 0 accumulation step
        success = self.accumulate()

        self.save_after_first_accumulation()

        if success:
            return

        logging.warning(
            "[red]Protocol 2 failed, protocol 3 is going to start",
            extra={"markup": True},
        )
        ask_about_protocol3(
            self.session.protocol3_action, self.session.number_of_error_frames
        )

        with self.session.new_timer("Protocol 3 pre-training"):
            self.pretrain()

        with self.session.new_timer("Protocol 3 accumulation"):
            for self.session.accumulation_trial in range(
                1, conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS + 1
            ):
                try:
                    self.accumulation_parachute_init(self.session.accumulation_trial)
                except IndexError:
                    logging.warning(
                        "There are no more Global Fragments to start new accumulations"
                    )
                    break

                success = self.accumulate()
                self.save_and_update_accumulation_parameters_in_parachute()
                if success:
                    logging.info("Accumulation after protocol 3 has been successful")
                    break
                logging.warning("Accumulation after protocol 3 failed")
            else:
                logging.warning(
                    "All accumulation trials after after Protocol 3 pretrain failed"
                )

            self.load_best_accumulation()

    def accumulate(self) -> bool:
        while self.accumulation_manager.new_global_fragments_for_training:
            early_stopped = perform_one_accumulation_step(
                self.accumulation_manager,
                self.session,
                self.identification_model,
                self.accumulation_network_params,
            )
            if early_stopped:
                logging.info("We don't need to accumulate more images")
                break
        else:
            logging.info("No more new images to accumulate")

        if (
            self.accumulation_manager.ratio_accumulated_images
            > conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
        ):
            logging.info("We accumulated enough images")
            return True
        logging.info("[red]We did not accumulate enough images", extra={"markup": True})
        return False

    def save_after_first_accumulation(self):
        """Set flags and save data"""
        logging.info("Saving first accumulation parameters")

        # if not self.restoring_first_accumulation:
        self.session.ratio_accumulated_images = (
            self.accumulation_manager.ratio_accumulated_images
        )
        self.session.percentage_of_accumulated_images = [
            self.session.ratio_accumulated_images
        ]
        self.session.save()
        self.list_of_fragments.save(self.session.fragments_path)
        self.list_of_fragments.save(self.session.accumulation_folder)
        self.list_of_global_fragments.save(self.session.global_fragments_path)

    def pretrain(self):
        create_dir(self.session.pretraining_folder, remove_existing=True)

        pretrain_network_params = NetworkParams(
            n_classes=self.session.n_animals,
            save_folder=self.session.pretraining_folder,
            model_name="identification_network",
            image_size=self.session.id_image_size,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
            knowledge_transfer_folder=self.session.knowledge_transfer_folder,
        )
        pretrain_network_params.save()

        # Initialize network
        if pretrain_network_params.knowledge_transfer_folder:
            self.identification_model = LearnerClassification.load_model(
                pretrain_network_params, knowledge_transfer=True
            )
            self.identification_model.fully_connected_reinitialization()
        else:
            self.identification_model = CNN.from_network_params(
                pretrain_network_params
            ).to(DEVICE)

        self.list_of_fragments.reset(roll_back_to="fragmentation")
        self.list_of_global_fragments.sort_by_distance_travelled()

        pretraining_counter = -1
        ratio_of_pretrained_images = 0.0
        while ratio_of_pretrained_images < conf.MAX_RATIO_OF_PRETRAINED_IMAGES:
            pretraining_counter += 1
            logging.info(
                "[bold]New pretraining iteration[/], using the #%s global fragment",
                pretraining_counter,
                extra={"markup": True},
            )
            pretrain_global_fragment(
                self.identification_model,
                pretrain_network_params,
                self.list_of_global_fragments.global_fragments[pretraining_counter],
                self.session.id_images_file_paths,
            )
            ratio_of_pretrained_images = (
                self.list_of_fragments.ratio_of_images_used_for_pretraining
            )

            logging.debug(
                f"{ratio_of_pretrained_images:.2%} of the images have been used during"
                " pretraining (if higher than"
                f" {conf.MAX_RATIO_OF_PRETRAINED_IMAGES:.2%} we stop pretraining)"
            )

    def accumulation_parachute_init(self, iteration_number: int) -> None:
        logging.info("Starting parachute accumulation %i", iteration_number)
        logging.info(
            "Setting #%d global fragment for accumulation", iteration_number - 1
        )

        self.list_of_global_fragments.sort_by_distance_travelled()
        first_global_fragment = self.list_of_global_fragments.global_fragments[
            iteration_number - 1
        ]

        create_dir(self.session.accumulation_folder, remove_existing=True)
        self.list_of_fragments.reset(roll_back_to="fragmentation")

        self.session.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
        )

        identify_first_global_fragment_for_accumulation(
            first_global_fragment,
            self.session,
            (
                LearnerClassification.load_model(self.accumulation_network_params)
                if self.session.identity_transfer
                else None
            ),
        )
        self.session.identities_groups = self.list_of_fragments.build_exclusive_rois()

        # Sort global fragments by distance
        self.list_of_global_fragments.sort_by_distance_to_the_frame(
            self.session.first_frame_first_global_fragment[iteration_number - 1]
        )
        logging.warning(
            "first_frame_first_global_fragment %s",
            self.session.first_frame_first_global_fragment,
        )
        logging.info(
            "We will restore the network from a previous pretraining: %s",
            self.session.pretraining_folder,
        )

        # Set saving folders
        self.accumulation_network_params.save_folder = self.session.accumulation_folder

        # Set restoring model_file
        self.accumulation_network_params.restore_folder = (
            self.session.pretraining_folder
        )

        self.identification_model = LearnerClassification.load_model(
            self.accumulation_network_params
        )

        self.identification_model.fully_connected_reinitialization()

        # Instantiate accumualtion manager
        self.accumulation_manager = AccumulationManager(
            self.session.n_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        logging.info("Start accumulation")

    def save_and_update_accumulation_parameters_in_parachute(self) -> None:
        logging.info(
            "Accumulated images"
            f" {self.accumulation_manager.ratio_accumulated_images:.2%}"
        )
        self.session.ratio_accumulated_images = (
            self.accumulation_manager.ratio_accumulated_images
        )
        self.session.percentage_of_accumulated_images.append(
            self.session.ratio_accumulated_images
        )
        self.list_of_fragments.save(
            self.session.accumulation_folder / "list_of_fragments.json"
        )

    def load_best_accumulation(self) -> None:
        logging.info("Saving second accumulation parameters")

        # Choose best accumulation
        self.session.accumulation_trial = int(
            np.argmax(self.session.percentage_of_accumulated_images)
        )
        logging.info(f"Best accumulation is #{self.session.accumulation_trial}")

        # Update ratio of accumulated images and  accumulation folder
        self.session.ratio_accumulated_images = (
            self.session.percentage_of_accumulated_images[
                self.session.accumulation_trial
            ]
        )

        # Load list of fragments of the best accumulation
        self.list_of_fragments = ListOfFragments.load(
            self.session.accumulation_folder / "list_of_fragments.json"
        )
        self.list_of_fragments.save(self.session.fragments_path)
        self.list_of_global_fragments.save(self.session.global_fragments_path)

        # set restoring folder
        logging.info("Restoring networks to best second accumulation")
        self.accumulation_network_params.restore_folder = (
            self.session.accumulation_folder
        )

        # Load pretrained network
        self.identification_model = LearnerClassification.load_model(
            self.accumulation_network_params
        )

        self.session.save()


def ask_about_protocol3(protocol3_action: str, n_error_frames: int) -> None:
    """Raises a IdtrackeraiError if protocol3_action is abort or aks and user answers abortion"""
    logging.info("Protocol 3 action: %s", protocol3_action)

    if protocol3_action == "abort":
        raise IdtrackeraiError(
            "Protocol 3 was going to start but PROTOCOL3_ACTION is set to 'abort'"
        )
    if protocol3_action == "continue":
        return

    if protocol3_action != "ask":
        raise ValueError(
            f'PROTOCOL3_ACTION "{protocol3_action}" not in ("ask", "abort", "continue")'
        )

    if n_error_frames > 0:
        logging.info(
            "Protocol 3 is a very time consuming algorithm and, in most cases, it"
            " can be avoided by redefining the segmentation parameters. As"
            " [red]there are %d frames with more blobs than animals[/red], we"
            " recommend you to abort the tracking session now and go back to the"
            " Segmentation app focusing on not having reflections, shades, etc."
            " detected as blobs. Check the following general recommendations:\n   "
            " - Define a region of interest to exclude undesired noise blobs\n    -"
            " Shrink the intensity (or background difference) thresholds\n    -"
            " Toggle the use of the background subtraction\n    - Shrink the blob's"
            " area thresholds",
            n_error_frames,
            extra={"markup": True},
        )
    else:
        logging.info(
            "Protocol 3 is a very time consuming algorithm and, in most cases, it"
            " can be avoided by redefining the segmentation parameters. As"
            " [bold]there are NOT frames with more blobs than animals[/bold], the"
            " video is unlikely to have non-animal blobs. Even so, you can choose"
            " to abort the tracking session and redefine the segmentation"
            " parameters (specially shrinking the intensity (or background"
            " difference) thresholds) or to continue with Protocol 3.",
            extra={"markup": True},
        )

    abort = None
    valid_answers = {"abort": True, "a": True, "continue": False, "c": False}
    while abort is None:
        answer_str = input(
            "What do you want to do now? Abort [A] or Continue [C]? "
        ).lower()
        if answer_str not in valid_answers:
            logging.warning("Invalid answer")
            continue
        abort = valid_answers[answer_str]
        logging.info("Answer --> Abort? %s", abort)
    if abort:
        raise IdtrackeraiError(
            "This is not an actual error: protocol 3 was going to start"
            " but PROTOCOL3_ACTION is set to 'ask' and used aborted."
        )
    return



import numpy as np


def split_data_train_and_validation(
    images: np.ndarray, labels: np.ndarray, validation_proportion: float, n_animals: int
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Splits a set of `images` and `labels` into training and validation sets

    Parameters
    ----------
    number_of_animals : int
        Number of classes in the set of images
    images : list
        List of images (arrays of shape [height, width])
    labels : list
        List of integers from 0 to `number_of_animals` - 1
    validation_proportion : float
        The proportion of images that will be used to create the validation set.


    Returns
    -------
    training_dataset : <DataSet object>
        Object containing the images and labels for training
    validation_dataset : <DataSet object>
        Object containing the images and labels for validation

    See Also
    --------
    :class:`get_data.DataSet`
    :func:`get_data.duplicate_PCA_images`
    """
    assert len(images) == len(labels)

    # shuffle images and labels
    shuffled_order = np.arange(len(images))
    np.random.shuffle(shuffled_order)
    images = images[shuffled_order]
    labels = labels[shuffled_order]

    # Init variables
    train_images = []
    train_labels = []
    validation_images = []
    validation_labels = []

    for i in np.unique(labels):
        # Get images of this individual
        individual_indices = labels == i
        this_indiv_images = images[individual_indices]
        this_indiv_labels = labels[individual_indices]

        n_images_validation = int(validation_proportion * len(this_indiv_labels) + 0.99)

        validation_images.append(this_indiv_images[:n_images_validation])
        validation_labels.append(this_indiv_labels[:n_images_validation])
        train_images.append(this_indiv_images[n_images_validation:])
        train_labels.append(this_indiv_labels[n_images_validation:])

    train_labels = np.concatenate(train_labels)

    train_weights = 1.0 - np.bincount(train_labels, minlength=n_animals) / len(
        train_labels
    )

    return (
        np.concatenate(train_images),
        train_labels,
        train_weights,
        np.concatenate(validation_images),
        np.concatenate(validation_labels),
    )



import logging
from pathlib import Path
from shutil import copyfile

import torch
from torch.nn import CrossEntropyLoss
from torch.optim import SGD, Adam
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import Session
from idtrackerai.utils import conf, load_id_images

from ..network import (
    CNN,
    DEVICE,
    LearnerClassification,
    NetworkParams,
    StopTraining,
    evaluate_only_acc,
    get_dataloader,
    get_onthefly_dataloader,
    train_loop,
)
from .accumulation_manager import (
    AccumulationManager,
    get_predictions_of_candidates_fragments,
)
from .identity_dataset import split_data_train_and_validation


def perform_one_accumulation_step(
    accumulation_manager: AccumulationManager,
    session: Session,
    identification_model: CNN,
    network_params: NetworkParams,
) -> bool:
    logging.info(
        f"[bold]Performing new accumulation, step {accumulation_manager.current_step}",
        extra={"markup": True},
    )
    id_img_paths = session.id_images_file_paths

    # Get images for training
    accumulation_manager.get_new_images_and_labels()

    # get a mixture of old and new images to train
    images_for_training, labels_for_training = (
        accumulation_manager.get_old_and_new_images()
    )

    (
        train_images,
        train_labels,
        train_weights,
        validation_images,
        validation_labels,
    ) = split_data_train_and_validation(
        load_id_images(id_img_paths, images_for_training),
        labels_for_training,
        conf.VALIDATION_PROPORTION,
        session.n_animals,
    )
    assert len(images_for_training) == len(labels_for_training)
    assert len(validation_images) > 0

    train_loader = get_dataloader(
        "training", train_images, train_labels, conf.BATCH_SIZE_IDCNN
    )
    val_loader = get_dataloader("validation", validation_images, validation_labels)

    criterion = CrossEntropyLoss(
        weight=torch.tensor(train_weights, dtype=torch.float32)
    ).to(DEVICE)

    if network_params.optimizer == "Adam":
        optimizer = Adam(identification_model.parameters(), **network_params.optim_args)
    elif network_params.optimizer == "SGD":
        optimizer = SGD(identification_model.parameters(), **network_params.optim_args)
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    learner = LearnerClassification(
        identification_model, criterion, optimizer, scheduler
    )

    stopping = StopTraining(
        epochs_limit=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
        overfitting_limit=(
            conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN_FIRST_ACCUM
            if accumulation_manager.current_step == 0
            else conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN
        ),
        plateau_limit=conf.LEARNING_RATIO_DIFFERENCE_IDCNN,
    )

    train_loop(learner, train_loader, val_loader, stopping)

    # free some RAM
    del train_loader, val_loader, train_images, validation_images

    accumulation_manager.update_fragments_used_for_training()
    accumulation_manager.update_used_images_and_labels()
    accumulation_manager.assign_identities_to_fragments_used_for_training()

    # compute ratio of accumulated images and stop if it is above random
    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    test_acc = test_model(accumulation_manager, id_img_paths, learner.model)

    # keep a copy of the penultimate model
    network_params.penultimate_model_path.unlink(missing_ok=True)
    if network_params.model_path.is_file():
        copyfile(network_params.model_path, network_params.penultimate_model_path)
    learner.save_model(
        network_params.model_path,
        test_acc=test_acc,
        ratio_accumulated=accumulation_manager.ratio_accumulated_images,
    )

    if (
        accumulation_manager.ratio_accumulated_images
        > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
    ):
        logging.info(
            "The ratio of accumulated images is higher than"
            f" {conf.THRESHOLD_EARLY_STOP_ACCUMULATION:.1%}, [bold]stopping"
            " accumulation by early stopping criteria",
            extra={"markup": True},
        )
        return True

    # Set accumulation parameters for rest of the accumulation
    # take images from global fragments not used in training (in the remainder test global fragments)
    if any(
        not global_fragment.used_for_training
        for global_fragment in accumulation_manager.list_of_global_fragments
    ):
        logging.info(
            "Generating [bold]predictions[/bold] on remaining global fragments",
            extra={"markup": True},
        )
        (
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_fragments_identifiers,
        ) = get_predictions_of_candidates_fragments(
            identification_model, id_img_paths, accumulation_manager.list_of_fragments
        )

        accumulation_manager.split_predictions_after_network_assignment(
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_fragments_identifiers,
        )

        accumulation_manager.assign_identities(session.accumulation_trial)
        accumulation_manager.update_accumulation_statistics()
        accumulation_manager.current_step += 1

    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    while len(session.accumulation_statistics_data) <= session.accumulation_trial:
        session.accumulation_statistics_data.append({})

    session.accumulation_statistics_data[session.accumulation_trial] = (
        accumulation_manager.accumulation_statistics
    )
    return False


def test_model(
    accumulation_manager: AccumulationManager, id_img_paths: list[Path], model: CNN
):
    """Takes a sample of the accumulated images to test model's accuracy"""
    logging.info(
        "Using a sample of all accumulated images to test model's overall accuracy"
    )
    test_images, test_labels = accumulation_manager.get_old_images()
    test_dataloader = get_onthefly_dataloader(test_images, id_img_paths, test_labels)
    test_acc = evaluate_only_acc(test_dataloader, model)
    logging.info(f"Current model has an overall accuracy of {test_acc:.3%}")
    return test_acc
